/* 
 * Auth: Christian Hansen (chansen@cs.umd.edu)
 *       University of Maryland, College Park
 * Desc: Functionality for manipulating irregular block decompositions.
 *       This code was adapted in part for Jae-Yong Lee's Meta-Chaos work.
 */

#include "birreg.h"
#include "addition.h"
#include "hash.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>

/* allocates a descriptor for an array block */
decomp_block_t* alloc_block(int rank) {
  decomp_block_t* block;  

  block = (decomp_block_t*)malloc(sizeof(decomp_block_t));
  block->rank = rank;
  block->size = (int*)malloc(rank*sizeof(int));
  block->globalLowerBound = (int*)malloc(rank*sizeof(int));
  block->globalUpperBound = (int*)malloc(rank*sizeof(int));
  block->coord = (int*)malloc(rank*sizeof(int));

  return block;
}

/* deallocates a descriptor for an array block */
void free_block(decomp_block_t* block) {
  if (block) {
    if (block->size) free(block->size);
    if (block->globalLowerBound) free(block->globalLowerBound);
    if (block->globalUpperBound) free(block->globalUpperBound);
    if (block->coord) free(block->coord);  
    free(block);
  }
}

/* prints array block descriptor information */
void print_block(decomp_block_t* block, int rank, char* indent) {
  int i;
  
  printf("%s<block>\n", indent);
  printf("%stask = %d\n", indent, block->task);  
  printf("%ssize =", indent);  
  for (i = 0; i < rank; i++) 
    printf(" %d", block->size[i]);
  printf("\n");
  printf("%slower = (", indent);
  for (i = 0; i < rank; i++) 
    printf(" %d", block->globalLowerBound[i]);
  printf(" )\n");
  printf("%supper = (", indent);
  for (i = 0; i < rank; i++) 
    printf(" %d", block->globalUpperBound[i]);
  printf(" )\n");
  printf("%scoord = (", indent);
  for (i = 0; i < rank; i++) 
    printf(" %d", block->coord[i]);
  printf(" )\n");
  printf("%s</block>\n", indent);
}

/* converts a decomposition tree into a list of block descriptors */
int store_tree(decomp_tree_t* tree, decomp_block_t** blocks, int pos) {
  int i;
  int res = 0;
  int total = 0;
  
  if (tree->leaf) {
    blocks[pos] = tree->leaf;
    return 1;
  } else {
    for (i = 0; i < tree->partitions; i++) {
      res = store_tree(&tree->children[i], blocks, pos);
      pos += res;
      total += res;
    }  
    return total;
  }
}

/* deallocates a decomposition tree */
void free_tree(decomp_tree_t* tree) {
  int i;
  
  if (tree) {
    if (tree->children) {
      for (i = 0; i < tree->partitions; ++i)
	free_tree(&(tree->children[i]));
      free(tree->children);
    }
    if (tree->keys) free(tree->keys);
    if (tree->leaf) free_block(tree->leaf);
  }
}

/* counts the number of blocks in a decomposition tree */
int count_tree(decomp_tree_t* tree) {
  int i;
  int res = 0;

  if (tree->leaf) {
    return 1;
  } else {
    for (i = 0; i < tree->partitions; i++)
      res += count_tree(&(tree->children[i]));
    return res;
  }
}

/* prints the information for a decomposition tree */
void print_tree(decomp_tree_t* tree, int rank, char* ind) {
  int i;
  char indent[10];

  sprintf(indent, "%s  ", ind);
  printf("%s-PARTITION-\n", indent);
  printf("%sdimension = %d\n", indent, tree->dimension);
  printf("%spartitions = %d\n", indent, tree->partitions);
  if (tree->children) {
    for (i = 0; i < tree->partitions; i++) {
      printf("%sKEY[%d] = %d\n", indent, i, tree->keys[i]);
      print_tree(&(tree->children[i]), rank, indent);
    }
  }
  if (tree->leaf) {
    print_block(tree->leaf, rank, indent);
  }
}

/* deallocates a decompostion descriptor */
void parti_free_decomp(parti_decomp_t* decomp) {
  if (decomp->referenceCount-- == 0) {
    if (decomp->size) free(decomp->size);
    free_tree(&(decomp->root));
    if (decomp->temp) delete_XDECOMP(decomp->temp);
    free(decomp);
  }
}

/* counts the number of blocks in a decompostion descriptor */
int count_decomp(parti_decomp_t* decomp) {
  return count_tree(&(decomp->root));  
}

/* converts a decomposition descriptor to a list of block descriptors */
void store_decomp(parti_decomp_t* decomp, decomp_block_t** blocks) {
  store_tree(&decomp->root, blocks, 0);
}

/* prints a decomposition descriptor */
void print_decomp(parti_decomp_t* decomp) {
  int i;
  char indent[10];

  printf("-DECOMP-\n");
  for (i = 0; i < decomp->rank; ++i) {
    printf("size[%d] = %d\n", i, decomp->size[i]);
  }
  sprintf(indent, " ");
  print_tree(&(decomp->root), decomp->rank, indent);
}

/* retrieves a list of block descriptors for a decomposition tree */
void get_task_blocks(int task, decomp_tree_t* tree, decomp_block_t*** blocks, int* count) {
  int i;
  decomp_block_t** copy;

  if (tree->leaf && tree->leaf->task == task) {
    copy = *blocks;
    *blocks = (decomp_block_t**)malloc(sizeof(decomp_block_t*)*(*count+1));
    for (i = 0; i < *count; ++i)
      (*blocks)[i] = copy[i];
    free(copy);
    (*blocks)[i] = tree->leaf;
    ++(*count);
  }
  
  if (tree->children) {
    for (i = 0; i < tree->partitions; ++i) {
      get_task_blocks(task, &(tree->children[i]), blocks, count);
    }
  }  
}

/* retrieves a list of block descriptors based on a global region */
void get_region_blocks(parti_region_t* region, decomp_tree_t* tree, 
		       decomp_block_t*** blocks, int* count) {
  int i, lower, upper; 
  decomp_block_t** copy;
  
  if (tree->leaf) {
    copy = *blocks;
    *blocks = (decomp_block_t**)malloc(sizeof(decomp_block_t*)*(*count+1));
    for (i = 0; i < *count; ++i) 
      (*blocks)[i] = copy[i];
    free(copy);
    (*blocks)[i] = tree->leaf;
    ++(*count);
  }
  
  if (tree->children) {
    lower = 0;
    for (i = 0; i < tree->partitions; ++i) {
      upper = tree->keys[i];
      if (region->lower[tree->dimension] <= upper && region->upper[tree->dimension] >= lower) {
	get_region_blocks(region, &(tree->children[i]), blocks, count);
      }
      lower = tree->keys[i] + 1;
    }
  }
}

/* the following functions appear are part of the public interface
 * and appear in multiblockparti.h */

/* allocates a decomposition descriptor */
parti_decomp_t* parti_create_idecomp(int rank, int* size) {
  int i;
  parti_decomp_t* decomp;

  if (rank <= 0 || size == 0) {
    fprintf(stderr, "invalid parameters\n");
    return 0;
  }

  decomp = (parti_decomp_t*)malloc(sizeof(parti_decomp_t));
  decomp->rank = rank;
  decomp->size = (int*)malloc(sizeof(int)*rank);
  for (i = 0; i < rank; ++i) {
    decomp->size[i] = size[i];
  }
  memset(&(decomp->root), 0, sizeof(decomp_tree_t));
  decomp->referenceCount = 0;
  decomp->temp = 0;

  return decomp;
}

int subsection(parti_decomp_t* decomp, int* dim, int* blk, int* cnt, int** idx, int depth) {  
  int i, blocks = 0;
  
  if (depth < decomp->rank) {
    if (parti_partition(decomp, dim[depth], blk, cnt[depth], idx[depth]) < cnt[depth]) {
      return 0;
    }
    for (i = 0; i < cnt[depth]; ++i) {
      blk[depth] = i;
      blocks += subsection(decomp, dim, blk, cnt, idx, depth+1);
    }
  } else {
    blocks = 1;
  }
  
  return blocks;
}

int parti_section(parti_decomp_t* decomp, int* dim, int* cnt, int** idx) {
  int sts, *blk;

  if (decomp == 0 || dim == 0 || cnt == 0 || idx == 0) {
    fprintf(stderr, "invalid parameters\n");
    return 0;
  }

  blk = calloc(decomp->rank, sizeof(int));  
  sts = subsection(decomp, dim, blk, cnt, idx, 0);
  free(blk);

  return sts;
}

/* creates a partition within a decomposition */
int parti_partition(parti_decomp_t* decomp, int dim, int* blk, int cnt, int* idx) {
  int i, depth = 0;
  decomp_tree_t* tree;

  if (decomp == 0 || dim < 0 || blk == 0 || cnt <= 1 || idx == 0) {
    fprintf(stderr, "invalid parameters\n");
    return 0;
  }
  
  if (idx[cnt-1] != decomp->size[dim]) {
    fprintf(stderr, "incomplete partitioning for dimension %d\n", dim);
    return 0;
  }

  for (tree = &(decomp->root); tree->children; tree = &(tree->children[blk[depth-1]])) {
    if (tree->dimension == dim) {
      printf("this dimension is already partitioned!\n");
      return 0;
    }
    ++depth;
  }
  
  if (depth >= decomp->rank) {
    printf("no dimensions left to partition!\n");
    return 0;
  }
    
  tree->dimension = dim;
  tree->partitions = cnt;
  tree->leaf = 0;
  tree->keys = (int*)malloc(sizeof(int)*cnt);
  tree->children = (decomp_tree_t*)malloc(sizeof(decomp_tree_t)*cnt);
  for (i = 0; i < cnt; ++i) {
    tree->keys[i] = idx[i] - 1; /* users specify the upper bound of the partition */
    memset(&(tree->children[i]), 0, sizeof(decomp_tree_t));      
  }

  return cnt;
}

/* verifies the decomposition tree structure and creates the block descriptors */
int verify_decomp(parti_decomp_t* decomp, decomp_tree_t* tree, int* dim,
		  int* lower, int* upper, int* coord, int depth, int* task) {
  int i, sts = 1;
  
  /* create the block descriptor */
  if (tree->partitions == 0) {
    if (tree->leaf) {
      fprintf(stderr, "decomp already aligned\n");
      return 0;
    }
    tree->leaf = alloc_block(decomp->rank);
    /* TODO: replace this arbitrary round robin task assignment? */
    tree->leaf->task = (*task)++;
    if (tree->leaf->task > PARTI_numprocs()) {
      fprintf(stderr, "partition blocks (%d) exceed number of tasks (%d)\n", tree->leaf->task, PARTI_numprocs());
      return 0;
    }
    memset(tree->leaf->size, 0, decomp->rank*sizeof(int));
    for (i = 0; i < depth; ++i) {
      tree->leaf->size[dim[i]] = upper[i] - lower[i] + 1;
      tree->leaf->globalLowerBound[dim[i]] = lower[i];
      tree->leaf->globalUpperBound[dim[i]] = upper[i];      
      tree->leaf->coord[dim[i]] = coord[i];
    }
    /* account for unpartitioned dimensions */
    for (i = 0; i < decomp->rank; ++i) {
      if (tree->leaf->size[i] == 0) {
	tree->leaf->size[i] = decomp->size[i];
	tree->leaf->globalLowerBound[i] = 0;
	tree->leaf->globalUpperBound[i] = decomp->size[i] - 1;
	tree->leaf->coord[i] = 0;
      }
    }
    return 1;
  }

  /* set the block extents for this dimension and recurse */
  lower[depth] = 0;
  for (i = 0; i < tree->partitions; ++i) {
    upper[depth] = tree->keys[i];
    coord[depth] = i;
    dim[depth] = tree->dimension;
    sts *= verify_decomp(decomp, &(tree->children[i]), dim, lower, upper, coord, depth+1, task);
    lower[depth] = tree->keys[i] + 1;
  }
  
  return sts;
}

/* creates a distributed array descriptor using the provided decomposition */
parti_darray_t* parti_ialign(parti_decomp_t* decomp, int* igcl, int* igcu) {
  int i, sts, count = 0;
  parti_darray_t* darray;
  int *dim, *lower, *upper, *coord;
  decomp_block_t** blocks = 0;
  int task = 0;

  dim = (int*)malloc(sizeof(int)*decomp->rank);
  lower = (int*)malloc(sizeof(int)*decomp->rank);
  upper = (int*)malloc(sizeof(int)*decomp->rank);
  coord = (int*)malloc(sizeof(int)*decomp->rank);
  sts = verify_decomp(decomp, &(decomp->root), dim, lower, upper, coord, 0, &task);
  free(lower);
  free(upper);
  free(coord);
  free(dim);

  if (!sts)
    return 0;

  darray = (parti_darray_t*)malloc(sizeof(parti_darray_t));
  darray->lowerGhostCells = (int*)malloc(sizeof(int)*decomp->rank);
  darray->upperGhostCells = (int*)malloc(sizeof(int)*decomp->rank);
  for (i = 0; i < decomp->rank; ++i) {
    darray->lowerGhostCells[i] = igcl[i];
    darray->upperGhostCells[i] = igcu[i];
  }
  darray->decomp = decomp;
  darray->decomp->referenceCount++;
  get_task_blocks(PARTI_myproc(), &(decomp->root), &blocks, &count);
  if (count > 1)
    return 0;

  if (count < 1) {
    /* this task is not participating in this array */
    darray->block = 0;
  } else {
    darray->block = *blocks;
    free(blocks);
  }
  darray->referenceCount = 0;    
  
  return darray;
}

/* prints distributed array descriptor information */
void print_darray(parti_darray_t* darray) {
  int i;

  if (darray && darray->decomp) {
    printf("-DARRAY-\n");
    printf("lowerGhost = (");
    for (i = 0; i < darray->decomp->rank; ++i) {
      printf(" %d", darray->lowerGhostCells[i]);
    }
    printf(" )\n");
    printf("upperGhost = (");
    for (i = 0; i < darray->decomp->rank; ++i) {
      printf(" %d", darray->upperGhostCells[i]);
    }
    printf(" )\n");
    print_decomp(darray->decomp);
  }
}

/* deallocates a distributed array descriptor */
void parti_free_darray(parti_darray_t* darray) {
  /* clean out cached schedules involving this array */
  clear_exch_table(darray);
  clear_sub_table(darray);
  
  if (darray->referenceCount-- == 0) {
    if (darray->lowerGhostCells) free(darray->lowerGhostCells);
    if (darray->upperGhostCells) free(darray->upperGhostCells);
    if (darray->decomp) parti_free_decomp(darray->decomp);
    free(darray);
  }
}

/* provides the lower/upper bound for the local array block(s) */
int parti_local_loop_bound(parti_darray_t* darray, int dim, int idx, int stride) {
  int lower, upper;
  int loopBound;

  if (!darray->block)
    return PARTI_OUT_OF_RANGE;

  lower = darray->block->globalLowerBound[dim];
  upper = darray->block->globalUpperBound[dim];
  
  if (idx < lower) {
    loopBound = idx;
    while (loopBound < lower)
      loopBound += stride;
    if (loopBound > upper)
      return PARTI_OUT_OF_RANGE; /* there are no elements in this block */
  } else if (idx > upper) {
    loopBound = idx;
    while (loopBound > upper)
      loopBound -= stride;
    if (loopBound < lower)
      return PARTI_OUT_OF_RANGE;
  } else {
    loopBound = idx;
  }

  return (loopBound - lower + darray->lowerGhostCells[dim]);
}

/* provides the local array size */
int parti_local_size(parti_darray_t* darray, int* size) {
  int i, total = 1;

  if (!darray->block) {
    for (i = 0; i < darray->decomp->rank; ++i) {
      if (size) {
	size[i] = 0;
      }
    }
    total = 0;
  } else {
    for (i = 0; i < darray->decomp->rank; ++i) {
      if (size) {
	size[i] = darray->block->size[i] + darray->lowerGhostCells[i] + darray->upperGhostCells[i];
      }
      total *= darray->block->size[i] + darray->lowerGhostCells[i] + darray->upperGhostCells[i];
    }
  }
 
  return total;
}

/* provides the lower global bound for the local array block(s) */
int parti_global_lower_bound(parti_darray_t* darray, int dim) {
  if (!darray->block)
    return PARTI_OUT_OF_RANGE;
  else
    return darray->block->globalLowerBound[dim];
}

/* provides the upper global bound for the local array block(s) */
int parti_global_upper_bound(parti_darray_t* darray, int dim) {  
  if (!darray->block)
    return PARTI_OUT_OF_RANGE;
  else
    return darray->block->globalUpperBound[dim];
}

/* converts a global index to a local index */
int parti_global_to_local(parti_darray_t* darray, int idx, int dim, int ghost) {
  int lower, upper;
  
  if (!darray->block)
    return PARTI_OUT_OF_RANGE;
  
  if (ghost) {
    lower = darray->block->globalLowerBound[dim] - darray->lowerGhostCells[dim];
    upper = darray->block->globalUpperBound[dim] + darray->upperGhostCells[dim];
  } else {
    lower = darray->block->globalLowerBound[dim];
    upper = darray->block->globalUpperBound[dim];
  }
  
  /* out of range */
  if (idx < lower || idx > upper)
    return PARTI_OUT_OF_RANGE;

  if (ghost)
    return idx - lower;
  else
    return idx - lower + darray->lowerGhostCells[dim];
}

/* converts a local index to a global index */
int parti_local_to_global(parti_darray_t* darray, int idx, int dim, int ghost) {
  int lower, upper;

  if (!darray->block)
    return PARTI_OUT_OF_RANGE;
  
  if (ghost) {
    lower = 0;
    upper = darray->lowerGhostCells[dim] + darray->upperGhostCells[dim] + darray->block->size[dim] - 1;
  } else {
    lower = darray->lowerGhostCells[dim];
    upper = darray->lowerGhostCells[dim] + darray->block->size[dim] - 1;
  }
  
  /* out of range */
  if (idx < lower || idx > upper) 
    return PARTI_OUT_OF_RANGE;

  if (ghost)
    return idx + darray->block->globalLowerBound[dim] - darray->lowerGhostCells[dim];
  else
    return idx + darray->block->globalLowerBound[dim];
}
