/*
 * Auth: Alan Sussman (als@cs.umd.edu)
 *       University of Maryland, College Park
 * Desc: The original MultiBlockParti interface
 */

/* Added by Serge Koren for merge with P++ */
#include "bsparti.h"
#include "addition.h"
#include "birreg.h"
#include "decomp.h"
#include "hash.h"
#include "helper.h"
#include "port.h"
#include "utils.h"
#include <assert.h>
#include <stdlib.h>
#include <stdio.h>

/* globals */
int FFlag;

/****************************************************************************
 *
 * lalbnd(dArray, dim, start, stride ) 
 * 
 *  adjusts the bound of the array being traversed
 *  
 * 
 * Inputs: dArray  - distributed array 
 *         dim     - the dimension of dArray being traversed
 *         start   - start of range being traversed
 *         stride  - distance between elements, positive means counting up,
 *                   negative means counting down
 *                   (strides added 2/93 - als)
 * 
 * LIMITATIONS: 
 *        - in C loops must use either "<=" or ">=" to specify loop termination
 *          (i.e.  "for (i=start; i<=stop; ..)")
 *        
 ***************************************************************************/
int lalbnd(DARRAY* dArray, int dim, int start, int stride) {
  return parti_local_loop_bound(dArray, dim, start, stride);
}

/****************************************************************************
 *
 * laubnd(dArray, dim, start, stride ) 
 * 
 *  adjusts the bound of the array being traversed
 *  
 * 
 * Inputs: dArray  - distributed array 
 *         dim     - the dimension of dArray being traversed
 *         start   - start of range being traversed
 *         stride  - distance between elements, positive means counting up,
 *                   negative means counting down
 *                   (strides added 2/93 - als)
 * 
 * LIMITATIONS: 
 *        - in C loops must use either "<=" or ">=" to specify loop termination
 *          (i.e.  "for (i=start; i<=stop; ..)")
 *        
 ***************************************************************************/
int laubnd(DARRAY* dArray, int dim, int stop, int stride) {
  return parti_local_loop_bound(dArray, dim, stop, stride);
}

/************************************************************************
 *
 * align(decomp, numDims, arrayDims, sizes, int_gCells_l, int_gCells_r
 *       ext_gCells_l, ext_gCells_r, extra_flag, decompDims) 
 * 
 * 
 * Inputs: decomp    - decomposition to which the array is being aligned
 *         numDims   - the number of dimensions of array 
 *         arrayDims  - dimension of the src array to be aligned 
 *         sizes      - size of array
 *         int_gCells_l - number of internal ghost cells on left in each dim
 *         int_gCells_r - number of internal ghost cells on right in each dim
 *         ext_gCells_l - number of external ghost cells on left in each dim
 *         ext_gCells_r - number of external ghost cells on right in each dim
 *         extra_flag   -Where to put extra cells if length of dim is not    
 *                 exactly divisible by number of processors
 *                 0. Default which is same as option 4.
 *                 1. on the left most processor                       
 *                 2. on the right most processor                      
 *                 3. Split equally if odd the extra one on left       
 *                 4. Split equally if odd the extra one on right      
 *         decompDims - the decomp dimension to which the array dim is 
 *                      being aligned
 * 
 * Returns: a distributed array
 * 
 ***********************************************************************/
struct dArray_rec* xalign(struct decomp_rec* decomp, int numDims, 
			  int* arrayDims, int* sizes, 
			  int* int_gCells_l, int* int_gCells_r, 
			  int* ext_gCells_l, int* ext_gCells_r, 
			  int* extra_flag, int* decompDims) {
  struct dArray_rec *dArray;
  int i, arrayDim, decompDim;
  int temp_L, temp_G, temp_P, t_extra;
  int* block; 
  
  dArray = (struct dArray_rec*)pmalloc("struct dArray_rec", sizeof(struct dArray_rec));
  dArray->nDims = numDims;
  block = dArray->left_ghostCells = (int*)calloc(11*numDims, sizeof(int));
  dArray->right_ghostCells = block + numDims;
  dArray->dimVecG = block + numDims*2;
  dArray->dimVecL = block + numDims*3;
  dArray->dimVecL_L = block + numDims*4;
  dArray->dimVecL_R = block + numDims*5;
  dArray->g_index_low = block + numDims*6;
  dArray->g_index_hi = block + numDims*7;
  dArray->local_size = block + numDims*8;
  dArray->decompDim = block + numDims*9;
  dArray->decompPosn = block + numDims*10;
  dArray->dimDist = (char*)calloc(numDims, sizeof(char));
  dArray->decomp = decomp;
  decomp->referenceCount++;
  dArray->referenceCount = 0;
  
  gray(decomp, PARTI_myproc(), dArray->decompPosn);
  
  for (i = 0; i < numDims; i++) {
    arrayDim = arrayDims[i];
    decompDim = decompDims[i];
    
    dArray->dimVecG[arrayDim] = sizes[i];    
    dArray->left_ghostCells[arrayDim] = int_gCells_l[i];
    dArray->right_ghostCells[arrayDim] = int_gCells_r[i];
    dArray->decompDim[arrayDim] = decompDim;
    if (decompDim == -1 || decomp->dimDist[decompDim] == '*' ||	decomp->dimProc[decompDim] == 1) {
      dArray->dimDist[arrayDim] = '*';
      dArray->dimVecL[arrayDim] = dArray->dimVecG[arrayDim];
      dArray->dimVecL_L[arrayDim] = dArray->dimVecG[arrayDim];
      dArray->dimVecL_R[arrayDim] = dArray->dimVecG[arrayDim];
    } else {
      temp_G = dArray->dimVecG[arrayDim] - ext_gCells_l[arrayDim] - ext_gCells_r[arrayDim];
      temp_P = decomp->dimProc[decompDim];
      temp_L = temp_G/temp_P;
      
      if (temp_L * temp_P < temp_G) {
	if (abs(temp_L*temp_P - temp_G) > abs((temp_L+1)*temp_P - temp_G))
	  temp_L++;
	t_extra = temp_G - temp_L*temp_P;
	dArray->dimVecL[arrayDim] = temp_L;
	dArray->dimVecL_L[arrayDim] = temp_L;
	dArray->dimVecL_R[arrayDim] = temp_L;
	
	switch (extra_flag[i]) {
	case 0:
	case 4:
	  dArray->dimVecL_L[arrayDim] = temp_L + t_extra/2;
	  dArray->dimVecL_R[arrayDim] = temp_L + t_extra - t_extra/2;
	  break;
	case 1:
	  dArray->dimVecL_L[arrayDim] = temp_L + t_extra;
	  break;
	case 2:
	  dArray->dimVecL_R[arrayDim] = temp_L + t_extra;
	  break;
	case 3:
	  dArray->dimVecL_R[arrayDim] = temp_L + t_extra/2;
	  dArray->dimVecL_L[arrayDim] = temp_L + t_extra - t_extra/2;
	  break;
	default:
	  fprintf(stderr, "Out of range value for extra_flag in align");
	  break;
	}
	
	if (dArray->dimVecL_R[arrayDim] <= 0 || dArray->dimVecL_L[arrayDim] <= 0) {
	  /* shouldn't have added 1 to temp_L in the first place */
	  temp_L--;	  
	  t_extra = temp_G - temp_L*temp_P;
	  dArray->dimVecL[arrayDim] = temp_L;
	  dArray->dimVecL_L[arrayDim] = temp_L;
	  dArray->dimVecL_R[arrayDim] = temp_L;
	  
	  switch (extra_flag[i]) {
	  case 0:
	  case 4:
	    dArray->dimVecL_L[arrayDim] = temp_L + t_extra/2;
	    dArray->dimVecL_R[arrayDim] = temp_L + t_extra - t_extra/2;
	    break;
	  case 1:
	    dArray->dimVecL_L[arrayDim] = temp_L + t_extra;
	    break;
	  case 2:
	    dArray->dimVecL_R[arrayDim] = temp_L + t_extra;
	    break;
	  case 3:
	    dArray->dimVecL_R[arrayDim] = temp_L + t_extra/2;
	    dArray->dimVecL_L[arrayDim] = temp_L + t_extra - t_extra/2;
	    break;
	  default:
	    break;
	  }
	}
      } else {
	dArray->dimVecL[arrayDim] = temp_L;
	dArray->dimVecL_R[arrayDim] = temp_L;
	dArray->dimVecL_L[arrayDim] = temp_L;
      }
      dArray->dimVecL_R[arrayDim] += ext_gCells_r[arrayDim];
      dArray->dimVecL_L[arrayDim] += ext_gCells_l[arrayDim];
      dArray->dimDist[arrayDim] = decomp->dimDist[decompDim];
    }
    
    /* Set the global bounds for my processor */
    if (active(PARTI_myproc(), decomp)) {
      /* Set the lower bound */
      if (decompDim < 0 || dArray->decompPosn[decompDim] == 0) {
	/* My processor is the first processor */
	dArray->g_index_low[arrayDim] = 0;
      } else {
	dArray->g_index_low[arrayDim] = dArray->dimVecL_L[arrayDim] + 
	  (dArray->decompPosn[decompDim]-1)*dArray->dimVecL[arrayDim];
      }
      /* Set the upper bound */
      if (decompDim < 0 || decomp->dimProc[decompDim] == 1) {
	/* only one processor is assigned to this decomp dimension, so */
	/* this processor is both the first and last */
	dArray->g_index_hi[arrayDim] = dArray->dimVecL_L[arrayDim] - 1;
      } else if (dArray->decompPosn[decompDim] == decomp->dimProc[decompDim]-1) {
	/* My processor is the last processor */
	dArray->g_index_hi[arrayDim] = dArray->dimVecL_L[arrayDim] + 
	  dArray->dimVecL_R[arrayDim] + (dArray->decompPosn[decompDim]-1)*dArray->dimVecL[arrayDim] - 1;
      } else {
	dArray->g_index_hi[arrayDim] = dArray->dimVecL_L[arrayDim] + 
	  (dArray->decompPosn[decompDim])*dArray->dimVecL[arrayDim] - 1;
      }
      dArray->local_size[arrayDim] = dArray->g_index_hi[arrayDim] - dArray->g_index_low[arrayDim] + 1;
    } else {
      /* this processor is not in the decomposition, so owns none of the darray */
      dArray->g_index_low[arrayDim] = -1;
      dArray->g_index_hi[arrayDim] = -1;
      dArray->local_size[arrayDim] = 0;
    }
  }
  
  return dArray;
}

/* builds the decomposition tree for a regular block decomposition */
void build_tree(decomp_tree_t* tree, struct dArray_rec* dad, int dim, int order, 
		int* sizes, int* lower, int* upper, int* coord) {
  int i, j;
  int* size_new;
  int* lower_new;
  int* upper_new;
  int* coord_new;
  int tmp;
  int own;
  
  if (dim == dad->decomp->nDims) { /* block node */
    tree->partitions = 0;
    tree->keys = 0;
    tree->children = 0;    
    tree->leaf = alloc_block(dad->decomp->nDims);
    for (i = 0; i < dad->decomp->nDims; i++) {
      tree->leaf->size[i] = sizes[i];
      tree->leaf->globalLowerBound[i] = lower[i];
      tree->leaf->globalUpperBound[i] = upper[i];	
      tree->leaf->coord[i] = coord[i];
    }
    
    /* calculate ownership of this block node using coord and gray fucntion */
    tmp = 1;
    own = dad->decomp->baseProc;    
    for (i = 1; i < dad->decomp->nDims; i++)
      tmp *= dad->decomp->dimProc[i];    
    for (i = 0; i < dad->decomp->nDims - 1; i++) {
      own += coord[i]*tmp;
      tmp /= dad->decomp->dimProc[i+1];
    }
    own += coord[i];
    tree->leaf->task = own;
    
  } else { /* nonblock node */ 
    tree->dimension = dad->decompDim[dim];
    tree->partitions = dad->decomp->dimProc[dim];
    tree->keys = (int*)calloc(tree->partitions, sizeof(int));
    tree->children = (decomp_tree_t*)calloc(tree->partitions, sizeof(decomp_tree_t));
    tree->leaf = 0;
    
    for (i = 0; i < tree->partitions; i++) {      

      /* size */
      size_new = (int*)calloc(dim+1, sizeof(int));      
      for (j = 0; j < dim; j++)
	size_new[j] = sizes[j];      
      if (i == 0)
	size_new[dim] = dad->dimVecL_L[dim];
      else if (i == tree->partitions - 1)
	size_new[dim] = dad->dimVecL_R[dim];
      else 
	size_new[dim] = dad->dimVecL[dim];      
      
      /* global start */
      lower_new = (int*)calloc(dim+1, sizeof(int));      
      for (j = 0; j < dim; j++)
	lower_new[j] = lower[j];      
      if (i == 0)
	lower_new[dim] = 0;      
      else    
	lower_new[dim] = dad->dimVecL_L[dim] + (i - 1)*(dad->dimVecL[dim]);
      
      /* global end */
      upper_new = (int*)calloc(dim+1, sizeof(int));      
      for (j = 0; j < dim; j++)
	upper_new[j] = upper[j];      
      if (i == (dad->decomp->dimProc[dim] - 1))
	upper_new[dim] = dad->dimVecG[dim] - 1;
      else if (i == (dad->decomp->dimProc[dim] - 2))
	upper_new[dim] = dad->dimVecG[dim] - 1 - dad->dimVecL_R[dim];
      else {
	upper_new[dim] = dad->dimVecG[dim] - 1 - dad->dimVecL_R[dim] 
	  - ((dad->decomp->dimProc[dim] - 2) - i)*dad->dimVecL[dim];
      }
      
      tree->keys[i] = upper_new[dim];

      /* coordinate */
      coord_new =  (int*)malloc((dim+1)*sizeof(int));
      for (j = 0; j < dim; j++)
	coord_new[j] = coord[j];      
      coord_new[dim] = i;      
      
      build_tree(&(tree->children[i]), dad, dim+1, i, size_new, lower_new, upper_new, coord_new);      
    }
  } 
   
  if (dim > 0) {
    free(sizes);
    free(lower);
    free(upper);
    free(coord);
  }
}

/* this will convert an old style block regular decomposition to the
 * new irregular block decomposition descriptor
 */
parti_darray_t* convert_darray(struct dArray_rec* dad, DECOMP* decomp) {
  parti_darray_t* darray;
  int i, count = 0; 
  decomp_block_t** blocks = 0;

  darray = (parti_darray_t*)malloc(sizeof(parti_darray_t));
  darray->lowerGhostCells = (int*)calloc(dad->nDims, sizeof(int));
  darray->upperGhostCells = (int*)calloc(dad->nDims, sizeof(int));
  for (i = 0; i < dad->nDims; ++i) {
    darray->lowerGhostCells[i] = dad->left_ghostCells[i];
    darray->upperGhostCells[i] = dad->right_ghostCells[i];
  }
  darray->decomp = decomp;
  darray->decomp->referenceCount++;
  build_tree(&(darray->decomp->root), dad, 0, 0, 0, 0, 0, 0);
  get_task_blocks(PARTI_myproc(), &(decomp->root), &blocks, &count);
  if (count > 1)
    return 0;

  if (count < 1) {
    /* this task is not participating in this array */
    darray->block = 0;
  } else {
    darray->block = *blocks;
    free(blocks);
  }
  darray->referenceCount = 0;
  
  return darray;
}

DARRAY* align(DECOMP* decomp, int numDims, 
	      int* arrayDims, int* sizes, 
	      int* int_gCells_l, int* int_gCells_r, 
	      int* ext_gCells_l, int* ext_gCells_r, 
	      int* extra_flag, int* decompDims) {
  struct dArray_rec *dArray;
  parti_darray_t* darray;

  if (decomp->temp) {
    dArray = xalign(decomp->temp, numDims, arrayDims, sizes, 
		    int_gCells_l, int_gCells_r, ext_gCells_l, ext_gCells_r, 
		    extra_flag, decompDims);
    darray = convert_darray(dArray, decomp);
    delete_XDARRAY(dArray);
    delete_XDECOMP(decomp->temp);
    decomp->temp = 0;
  } else {
    darray = parti_ialign(decomp, int_gCells_l, int_gCells_r);
  }
  
  return darray;
}

/*************************************************************************
 *
 * void laSizes (dArray, sizes)
 * 
 *  This procedure computes the local size in each dimension for an 
 *  array whose physical and distribution characteristic
 *  are described in the input argument  dArray.
 *  This routine is used for memory allocation.
 * 
 * Inputs: dArray -  pointer to the distributed array
 *         sizes  -  the local size for each dimension of array
 *                 
 * Returns: nothing
 * 
 ************************************************************************/
void laSizes(DARRAY* dArray, int* info) {
  int total;
  total = parti_local_size(dArray, info);
  return;
}

/*************************************************************************
 *
 *  int gLBnd (dArray, dim)
 * 
 *  This procedure computes and returns the lower bound of the range of
 *  locally stored global indices.  A description of the arrays physical 
 *  and distribution characteristics is passed as an input parameter 
 *  along with the dimension being queried.
 *
 * 
 * Inputs: dArray -  pointer to the distributed array
 *         dim    -  the dimension of array being queried
 *                 
 * Returns: an integer corresponding to the lower bound of the
 *          range of global indices stored locally.
 * 
 ************************************************************************/
int gLBnd(DARRAY* dArray, int dim) {
  return parti_global_lower_bound(dArray, dim);
}

/*************************************************************************
 *
 *  int gUBnd (dArray, dim)
 * 
 *  This procedure computes and returns the upper bound of the range of
 *  locally stored global indices.  A description of the arrays physical 
 *  and distribution characteristics is passed as an input parameter 
 *  along with the dimension being queried.
 *
 * 
 * Inputs: dArray -  pointer to the distributed array
 *         dim    -  the dimension of array being queried
 *                 
 * Returns: an integer corresponding to the upper bound of the
 *          range of global indices stored locally.
 * 
 ************************************************************************/
int gUBnd(DARRAY* dArray, int dim) {
  return parti_global_upper_bound(dArray, dim);
}

/*************************************************************************
 *
 *  int globalToLocal (dArray, index, dim)
 * 
 *
 * 
 * Inputs: dArray -  pointer to the distributed array
 *         index  -  the global index value whose local
 *                   value is being computed.
 *         dim    -  the array dimension of interest
 *                 
 * Returns: an integer representing the local index value, or -1 if the
 *          processor doesn't own that part of the darray 
 * 
 ************************************************************************/
int globalToLocal(DARRAY* dArray, int index, int dim) {
  return parti_global_to_local(dArray, index, dim, 0);
}

/*************************************************************************
 *
 *  int globalToLocalWithGhost (dArray, index, dim)
 * 
 *
 * 
 * Inputs: dArray -  pointer to the distributed array
 *         index  -  the global index value whose local
 *                   value is being computed.
 *         dim    -  the array dimension of interest
 *                 
 * Returns: an integer representing the local index value, or -1 if the
 *          processor doesn't own that part of the darray and doesn't have
 *          a copy of that global index value as an internal ghost cell
 * 
 ************************************************************************/
int globalToLocalWithGhost(DARRAY* dArray, int index, int dim) {
  return parti_global_to_local(dArray, index, dim, 1);
}

/*************************************************************************
 *
 *  int localToGlobal (dArray, index, dim)
 * 
 *
 * 
 * Inputs: dArray -  pointer to the distributed array
 *         index  -  the local index value whose global 
 *                   value is being computed.
 *         dim    -  the array dimension of interest
 *                 
 * Returns: an integer representing the global index value, or -1 if the
 *          index is not a real darray element (e.g. a ghost cell, or an
 *          out of bounds index)
 * 
 ************************************************************************/
int localToGlobal(DARRAY* dArray, int index, int dim) {
  return parti_local_to_global(dArray, index, dim, 0);
}

/*************************************************************************
 *
 *  int localToGlobalWithGhost (dArray, index, dim)
 * 
 *
 * 
 * Inputs: dArray -  pointer to the distributed array
 *         index  -  the local index value whose global 
 *                   value is being computed.
 *         dim    -  the array dimension of interest
 *                 
 * Returns: an integer representing the global index value, or < 0 if the
 *          index is not a real darray element or an internal ghost cell
 *          (e.g. an out of bounds index or an internal ghost cell that
 *           doesn't correspond to a real index)
 * 
 ************************************************************************/
int localToGlobalWithGhost(DARRAY* dArray, int index, int dim) {
  return parti_local_to_global(dArray, index, dim, 1);
}

/****************************************************************
 * remove all stored exchange schedules
 ****************************************************************/
void remove_exch_scheds() {
  destroy_exch_table();
}

/****************************************************************
 * remove all stored subarray_exch schedules
 ****************************************************************/
void remove_subarray_scheds() {
  destroy_sub_table();
}

/****************************************************************
 * frees the schedule, and removes it from the hash table where
 * it is saved
 ****************************************************************/
void parti_free_sched(parti_sched_t* sched) {
  if (sched == NULL)
    return;
  
  /* remove from the table of schedules (for Fortran) */
  remove_sch_table(sched);  
  
  delete_SCHED(sched);
}
