/*
 * Auth: Alan Sussman (als@cs.umd.edu)
 *       University of Maryland, College Park
 * Desc:
 */

#include "exchsched.h"
#include "decomp.h"
#include "hash.h"
#include "helper.h"
#include "utils.h"
#include <stdlib.h>
#include <assert.h>
#include <math.h>

#define min(A,B) ((A)<(B)?(A):(B))
#define max(A,B) ((A)>(B)?(A):(B))

SCHED* exchSched(DARRAY* dArrayPtr, int dim, int fill) {
  return parti_ghost_sched_abs(dArrayPtr, dim, fill);
}

SCHED* ghostFillSched(DARRAY* dArrayPtr, int ndims, int* fillVec) {
  return parti_ghost_sched_vector(dArrayPtr, fillVec);
}

SCHED* ghostFillSpanSched(DARRAY* dArrayPtr, int ndims, int* fillVec) {
  return parti_ghost_sched_span(dArrayPtr, fillVec);
}

SCHED* ghostFillAllSched(DARRAY* dArrayPtr) {
  return parti_ghost_sched_all(dArrayPtr);
}

/*************************************************************************
 *
 * merge_sched(total_sched, new_sched)
 *
 *  merges two schedules, putting the result into the first one.  Since in
 *  a schedule there can be only one send and one receive SchedData for
 *  each processor, the new schedule will overwrite a previous entry in
 *  total_sched for the same processor
 *
 * Inputs: total_sched - the schedule to be merged into
 *         new_sched   - the schedule that is being merged in
 *
 ************************************************************************/
void merge_sched(SCHED* total_sched, SCHED* new_sched) {
  int proc, i, o, p, l, m, u, v;
  SchedData *data, *new_data;

  if (new_sched == NULL) 
    return;
  
  for (proc = 0; proc < PARTI_numprocs(); proc++) {
    if (new_sched->rMsgSz[proc] > 0) {
/*       assert(new_sched->rData[proc]->proc == proc); */
      if (total_sched->rMsgSz[proc] > 0) {
/* 	printf("<<<MERGE RSCHED>>>\n"); */
/* 	print_sched(total_sched); */
/* 	print_sched(new_sched); */

	/* expand the receive schedule data */
	total_sched->rMsgSz[proc] += new_sched->rMsgSz[proc];
	assert(total_sched->rData[proc] != 0);
	data = total_sched->rData[proc];
	new_data = new_sched->rData[proc];
	assert(data->proc == new_data->proc);
	assert(data->numDims == new_data->numDims);
	
	o = data->startPosn;
	p = new_data->startPosn;
	data->startPosn = 0;
	for (i = 0; i < data->numDims; ++i) {
	  l = o/data->str[i];
	  m = p/new_data->str[i];
	  data->startPosn += min(l, m) * data->str[i];
	  u = l + data->numelem[i];
	  v = m + new_data->numelem[i];
	  data->numelem[i] = max(u, v) - min(l, m);
	  o %= data->str[i];
	  p %= new_data->str[i];
	}

      } else {
	/* copy the receive schedule data */
	total_sched->rMsgSz[proc] = new_sched->rMsgSz[proc];
	total_sched->rData[proc] = (SchedData*)malloc(sizeof(SchedData));
	data = total_sched->rData[proc];
	new_data = new_sched->rData[proc];
	data->proc = new_data->proc;
	data->numDims = new_data->numDims;
	data->startPosn = new_data->startPosn;
	for (i = 0; i < data->numDims; i++) {
	  data->numelem[i] = new_data->numelem[i];
	  data->str[i] = new_data->str[i];
	}
      }
    }
    
    if (new_sched->sMsgSz[proc] > 0) {
/*       assert(new_sched->sData[proc]->proc == proc); */
      if (total_sched->sMsgSz[proc] > 0) {
/* 	printf("<<<MERGE SSCHED>>>\n"); */
/* 	print_sched(total_sched); */
/* 	print_sched(new_sched); */

	/* expand the send schedule data */
	total_sched->sMsgSz[proc] += new_sched->sMsgSz[proc];
	assert(total_sched->sData[proc] != 0);
	data = total_sched->sData[proc];
	new_data = new_sched->sData[proc];
	assert(data->proc == new_data->proc);
	assert(data->numDims == new_data->numDims);
	
	o = data->startPosn;
	p = new_data->startPosn;
	data->startPosn = 0;
	for (i = 0; i < data->numDims; ++i) {
	  l = o/data->str[i];
	  m = p/new_data->str[i];
	  data->startPosn += min(l, m) * data->str[i];
	  u = l + data->numelem[i];
	  v = m + new_data->numelem[i];
	  data->numelem[i] = max(u, v) - min(l, m);
	  o %= data->str[i];
	  p %= new_data->str[i];
	}

      } else {
	/* copy the send schedule data */
	total_sched->sMsgSz[proc] = new_sched->sMsgSz[proc];
	total_sched->sData[proc] = (SchedData*)malloc(sizeof(SchedData));
	data = total_sched->sData[proc];
	new_data = new_sched->sData[proc];
	data->proc = new_data->proc;
	data->numDims = new_data->numDims;
	data->startPosn = new_data->startPosn;
	for (i = 0; i < data->numDims; i++) {
	  data->numelem[i] = new_data->numelem[i];
	  data->str[i] = new_data->str[i];
	}
      }
    }
  }
}

/* 
   compute all the combinations of n choose p items, recursively
   
   Assumes n > 0, p > 0 and n >= p
   
   The representation of the combination is a vector of 0's and 1's of
   length n, where 1 means the elements is in the set
   
   partial: the vector of elements selected so far
   place  : the position in the vector currently being looked at
   comb_list: the list of all generated combinations
*/
void comb(int n, int p, int* partial, int place, list comb_list) {
  int *new_partial, i;
  
  if (n - place <= p) {
    /* must add all remaining elements */
    for (i = place; i < n; i++) {
      partial[i] = 1;
    }
    insert_list(partial, comb_list);
    return;
  }
  
  if (p == 0) {
    /* have all the elements needed */
    for (i = place; i < n; i++) {
      partial[i] = 0;
    }
    insert_list(partial, comb_list);
    return;
  }
  
  /* recurse with and without the element in position "place" */
  new_partial = (int*)calloc(n, sizeof(int));
  for (i = 0; i < place; i++) {
    new_partial[i] = partial[i];
  }
  
  partial[place]     = 0;    /* not in partial */
  new_partial[place] = 1;    /* in new_partial */
  
  comb(n, p, partial, place+1, comb_list);  
  comb(n, p-1, new_partial, place+1, comb_list);
}

/* 
   compute all the permutations of n items, recursively, ignoring positions
   where the vector "vec" is 0, meaning there are no ghost cells in that
   dimension
   
   Assumes n > 0

   The representation of the permutation is a vector of 0's and non-zeros of
   length n, where a non-zero is a fill of that amount in that dimension
   of the darray
   
   vec: the vector with the number of ghost cells in each dimension, with 0
   meaning no ghost cells
   partial: the vector of elements selected so far
   curr_dim  : the position in the vector currently being looked at
   perm_list: the list of all generated permutations
*/
void perm(int len, int* lvec, int* rvec, int* partial, int pos, list perm_list) {
  int *new_partial_l, *new_partial_r;
  int i;
  
  if (pos >= len) {
    /* have a complete fill vector */
    insert_list(partial, perm_list);
    return;
  }
  
  /* recurse with all choices for the ghost cells in position "curr_dim" */
  new_partial_l = (int*)calloc(len, sizeof(int));
  new_partial_r = (int*)calloc(len, sizeof(int));
  if (new_partial_l == new_partial_r) { 
    abort(); 
  }
  for (i = 0; i < pos; ++i) {
    new_partial_l[i] = partial[i];
    new_partial_r[i] = partial[i];
  }
  partial[pos] = 0;
  new_partial_l[pos] = lvec[pos];
  new_partial_r[pos] = rvec[pos];
  perm(len, lvec, rvec, partial, pos+1, perm_list);
  if (new_partial_l[pos] == 0)
    free(new_partial_l);
  else
    perm(len, lvec, rvec, new_partial_l, pos+1, perm_list);
  if (new_partial_r[pos] == 0) 
    free(new_partial_r);
  else
    perm(len, lvec, rvec, new_partial_r, pos+1, perm_list);
}
