#include "reader.h"
#include <hdf5.h>
#include <assert.h>
#include <stdlib.h>

void* source_array(char* filename, char* setname, parti_darray_t* darray) {
  hid_t file, dataset, dataspace, memspace, datatype, typesize;
  hsize_t* size;
  hssize_t* setOffset, *memOffset, *setExtent, *memExtent;
  herr_t sts;
  int i, total = 1;
  int rank;
  void* data;
  
  if (darray->block == 0)
    return 0;

  /* open the file */
  file = H5Fopen(filename, H5F_ACC_RDONLY, H5P_DEFAULT);
  if (file < 0) {
    fprintf(stderr, "hdf5 file not found\n");
    return 0;
  }

  /* get the dataset */
  dataset = H5Dopen(file, setname);
  if (dataset < 0) {
    fprintf(stderr, "dataset not found\n");
    return 0;
  }

  /* get the dataspace */
  dataspace = H5Dget_space(dataset);
  if (dataspace < 0) {
    fprintf(stderr, "dataspace not found\n");
    return 0;
  }

  /* get the array rank */
  rank = H5Sget_simple_extent_ndims(dataspace);
  if (rank < 0) {
    fprintf(stderr, "error reading array rank\n");
    return 0;
  }

  assert(rank == darray->decomp->rank);

  /* get the array size */
  size = (hsize_t*)calloc(rank, sizeof(hsize_t));
  sts = H5Sget_simple_extent_dims(dataspace, size, NULL);
  if (sts < 0) {
    fprintf(stderr, "error reading array size\n");
    free(size);
    return 0;
  }  

  for (i = 0; i < rank; ++i) {
    assert((int)size[i] == darray->decomp->size[i]);
    size[i] = (hsize_t)(darray->block->size[i] + darray->lowerGhostCells[i] + darray->upperGhostCells[i]);
    total *= darray->block->size[i] + darray->lowerGhostCells[i] + darray->upperGhostCells[i];    
  }
  
  /* create a memory space */
  memspace = H5Screate_simple(rank, size, NULL);
  if (memspace < 0) {
    fprintf(stderr, "could not allocate memory space\n");
    free(size);
    return 0;
  }
  
  /* define the memory hyperslab */
  memOffset = (hssize_t*)calloc(rank, sizeof(hssize_t));
  memExtent = (hssize_t*)calloc(rank, sizeof(hssize_t));
  for (i = 0; i < rank; ++i) {
    memOffset[i] = (hssize_t)darray->lowerGhostCells[i];
    memExtent[i] = (hssize_t)darray->block->size[i];
  }

  H5Sselect_hyperslab(memspace, H5S_SELECT_SET, memOffset, NULL, memExtent, NULL);
  
  /* define the dataset hyperslab */
  setOffset = (hssize_t*)calloc(rank, sizeof(hssize_t));
  setExtent = (hssize_t*)calloc(rank, sizeof(hssize_t));
  for (i = 0; i < rank; ++i) {
    setOffset[i] = (hssize_t)darray->block->globalLowerBound[i];
    setExtent[i] = (hssize_t)darray->block->size[i];
  }  

  H5Sselect_hyperslab(dataspace, H5S_SELECT_SET, setOffset, NULL, setExtent, NULL);
  
  /* determine the element type */
  datatype = H5Dget_type(dataset);
  typesize = H5Tget_size(datatype);

  switch (H5Tget_class(datatype)) {
  case H5T_INTEGER:
    if ((int)typesize == sizeof(char)) {
      data = (void*)calloc(total, sizeof(char));
      memset(data, 0, total*sizeof(char));
      sts = H5Dread(dataset, H5T_NATIVE_CHAR, memspace, dataspace, H5P_DEFAULT, data);
    } else if ((int)typesize == sizeof(int)) {
      data = (void*)calloc(total, sizeof(int));
      memset(data, 0, total*sizeof(int));
      sts = H5Dread(dataset, H5T_NATIVE_INT, memspace, dataspace, H5P_DEFAULT, data);
    }
    break;
  case H5T_FLOAT:    
    if ((int)typesize == sizeof(float)) {
      data = (void*)calloc(total, sizeof(float));
      memset(data, 0, total*sizeof(float));
      sts = H5Dread(dataset, H5T_NATIVE_FLOAT, memspace, dataspace, H5P_DEFAULT, data);
    } else if ((int)typesize == sizeof(double)) {
      data = (void*)calloc(total, sizeof(double));
      memset(data, 0, total*sizeof(double));
      sts = H5Dread(dataset, H5T_NATIVE_DOUBLE, memspace, dataspace, H5P_DEFAULT, data);
    }
    break;
  }
  
  if (sts < 0) {
    fprintf(stderr, "error reading data\n");
    free(data);
    free(setOffset);
    free(setExtent);
    free(memOffset);
    free(memExtent);
    free(size);
    return 0;
  }

  free(setOffset);
  free(setExtent);
  free(memOffset);
  free(memExtent);
  free(size);
  
  H5Dclose(dataset);
  H5Sclose(dataspace);
  H5Sclose(memspace);
  H5Fclose(file);

  return data;
}
