/* 
 * Auth: Christian Hansen (chansen@cs.umd.edu)
 *       University of Maryland, College Park
 * Desc: Initializes a MBP decomposition with a XML specification
 */

#include "partitioner.h"
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <libxml/xmlmemory.h>
#include <libxml/parser.h>

int parse_partition(xmlDocPtr doc, xmlNodePtr curr, parti_decomp_t* decomp, int* blk, int depth) {
  xmlChar* attr;
  int i = 0, dim = 0, cnt = 0, *idx = 0, sts;

  if (!decomp)
    return 0;
  
  attr = xmlGetProp(curr, (const xmlChar*)"dimension"); 
  if (attr) {
    dim = atoi(attr); 
    xmlFree(attr);
  }

  attr = xmlGetProp(curr, (const xmlChar*)"count"); 
  if (attr) {
    cnt = atoi(attr); 
    xmlFree(attr);
  }

  if (depth > 0) {
    attr = xmlGetProp(curr, (const xmlChar*)"block");
    if (attr) {
      blk[depth-1] = atoi(attr);
      xmlFree(attr);
    }
  }

  idx = calloc(cnt, sizeof(int));
  curr = curr->xmlChildrenNode;
  while (curr &&  i < cnt) {
    if ((!xmlStrcmp(curr->name, (const xmlChar*)"index"))) {
      attr = xmlNodeListGetString(doc, curr->xmlChildrenNode, 1);
      if (attr) {
	idx[i] = atoi(attr);
	xmlFree(attr);
      }
      ++i;
    }
    curr = curr->next;
  }

  sts = parti_partition(decomp, dim, blk, cnt, idx);
  free(idx);
  if (sts < 2) {
    return 0;
  }

  while (curr) {
    if ((!xmlStrcmp(curr->name, (const xmlChar*)"partition"))) {
      if (!parse_partition(doc, curr, decomp, blk, depth+1))
	return 0;
    }
    curr = curr->next;
  }

  return 1;
}

parti_decomp_t* parse_document(char* filename) {
  xmlDocPtr doc;
  xmlNodePtr curr;  
  xmlChar* attr;

  int i, rank = 0, *size = 0, *blk = 0;
  parti_decomp_t* decomp = 0;
  char* p;

  doc = xmlParseFile(filename);
  if (!doc) {
    fprintf(stderr, "document not parsed successfully\n");
    return 0;
  }
  
  curr = xmlDocGetRootElement(doc);
  if (!curr) {
    fprintf(stderr, "found an empty document\n");
    xmlFreeDoc(doc);
    return 0;
  }
  
  if (xmlStrcmp(curr->name, (const xmlChar*)"decomposition")) {
    fprintf(stderr, "document of the wrong type\n");
    xmlFreeDoc(doc);
    return 0;
  }

  attr = xmlGetProp(curr, (const xmlChar*)"rank");
  if (attr) {
    rank = atoi(attr);
    xmlFree(attr);
  }
#if PARTI_DEBUG > 1
  printf("rank: %d\n", rank);
#endif
  
  attr = xmlGetProp(curr, (const xmlChar*)"size");
  if (attr) {
    size = calloc(rank, sizeof(int));
    p = attr;
    for (i = 0; i < rank; ++i) {
      size[i] = atoi(p);
      p = strchr(p, 'x');
      ++p;
    }
    xmlFree(attr);
  }

#if PARTI_DEBUG > 1
  printf("size: [ ");
  for (i = 0; i < rank; ++i)
    printf("%d ", size[i]);
  printf("]\n");
#endif
  
  blk = calloc(rank, sizeof(int));
  decomp = parti_create_idecomp(rank, size);
  for (curr = curr->xmlChildrenNode; curr; curr = curr->next) {
    if (!xmlStrcmp(curr->name, (const xmlChar*)"partition")) {
      if (!parse_partition(doc, curr, decomp, blk, 0)) {
	free(blk);
	free(size);
	xmlFreeDoc(doc);
	parti_free_decomp(decomp);
	return 0;
      }
    }
  }
  
  free(blk);
  free(size);
  xmlFreeDoc(doc);

  return decomp;
}
