/*
 * Auth: Alan Sussman (als@cs.umd.edu)
 *       Christian Hansen (chansen@cs.umd.edu)
 *       University of Maryland, College Park
 * Desc:
 */

#include "bsparti.h"
#include "hash.h"
#include "print.h"
#include "config.h"
#include <stdlib.h>

#ifdef FORTRAN

#define FVPROC_F77      F77_FUNC(fvproc, FVPROC)
#define FDECOMP_F77     F77_FUNC(fdecomp, FDECOMP)
#define FEMBED_F77      F77_FUNC(fembed, FEMBED)
#define FDISTRIBUTE_F77 F77_FUNC(fdistribute, FDISTRIBUTE)
#define IFSECTION_F77   F77_FUNC(ifsection, IFSECTION)
#define IFPARTITION_F77 F77_FUNC(ifpartition, IFPARTITION)
#define IFALIGN_F77     F77_FUNC(ifalign, IFALIGN)

#define IFLALBND_F77    F77_FUNC(iflalbnd, IFLALBND)
#define IFLAUBND_F77    F77_FUNC(iflaubnd, IFLAUBND)

#define IFGHOSTFILL_SCHED_F77     F77_FUNC_(ifghostfill_sched, IFGHOSTFILL_SCHED)
#define IFGHOSTFILLSPAN_SCHED_F77 F77_FUNC_(ifghostfillspan_sched, IFGHOSTFILLSPAN_SCHED)
#define IFGHOSTFILLALL_SCHED_F77  F77_FUNC_(ifghostfillall_sched, IFGHOSTFILLALL_SCHED)
#define IFEXCH_SCHED_F77          F77_FUNC_(ifexch_sched, IFEXCH_SCHED)
#define IFSUBARRAY_SCHED_F77      F77_FUNC_(ifsubarray_sched, IFSUBARRAY_SCHED)

#define FIDATA_MOVE_F77 F77_FUNC_(fidata_move, FIDATA_MOVE)
#define FFDATA_MOVE_F77 F77_FUNC_(ffdata_move, FFDATA_MOVE)
#define FDDATA_MOVE_F77 F77_FUNC_(fddata_move, FDDATA_MOVE)
#define FCDATA_MOVE_F77 F77_FUNC_(fcdata_move, FCDATA_MOVE)

#define FLASIZES_F77 F77_FUNC(flasizes, FLASIZES)

#define FFREE_SCHED_F77            F77_FUNC_(ffree_sched, FFREE_SCHED)
#define REMOVE_EXCH_SCHEDS_F77     F77_FUNC_(remove_exch_scheds, REMOVE_EXCH_SCHEDS)
#define REMOVE_SUBARRAY_SCHEDS_F77 F77_FUNC_(remove_subarray_scheds, REMOVE_SUBARRAY_SCHEDS)

#define IFGLBND_F77 F77_FUNC_(ifglbnd, IFGLBND)
#define IFGUBND_F77 F77_FUNC_(ifgubnd, IFGUBND)

#define IFGLOBAL_TO_LOCAL_F77            F77_FUNC_(ifglobal_to_local, IFGLOBAL_TO_LOCAL)
#define IFGLOBAL_TO_LOCAL_WITH_GHOST_F77 F77_FUNC_(ifglobal_to_local_with_ghost, IFGLOBAL_TO_LOCAL_WITH_GHOST)
#define IFLOCAL_TO_GLOBAL_F77            F77_FUNC_(iflocal_to_global, IFLOCAL_TO_GLOBAL)
#define IFLOCAL_TO_GLOBAL_WITH_GHOST_F77 F77_FUNC_(iflocal_to_global_with_ghost, IFLOCAL_TO_GLOBAL_WITH_GHOST)

#define FCLEANUP_F77 F77_FUNC(fcleanup, FCLEANUP)

/*****************************************************************
 * Parameters :
 *   vp       : Integer describing virtual processor set
 *   ndim     : Number of dimensions in VP set
 *            : (Current Limitation : must be 1 dim)
 *   sizes    : VP set size in each dimension
 * creates virtual processor space and initializes hash table for darrays
 *
 * Virtual Processor number in each dimension start at 1 (for FORTRAN)
 *******************************************************************/
void FVPROC_F77(int* vp, int* ndim, int* sizes) {
  FFlag = 1;
  *vp = insert_vproc_table(vProc(*ndim, sizes));
  FFlag = 0;
}

/***************************************************************
 * Parameters :
 *   decomp : return value of decomposition ( Integer )
 *   ndim   : Number of Dimensions in Decomposition
 *   dimInfo: Sizes of dimensions
 *
 * Creates a decomposition
 ****************************************************************/
void FDECOMP_F77(int* decomp, int* ndim, int* dimInfo) { 
  FFlag = 1;
  *decomp = insert_decomp_table(create_decomp(*ndim, dimInfo));
  FFlag = 0;
}

/***************************************************************
 * Parameters :
 *   decomp   : decomposition (Integer)
 *   vp       : Virtual processor set
 *   startPosn: Starting position in VP set
 *   endPosn  : End position in VP set
 *
 * Embeds a decomposition in subset of virtual processor space
 ****************************************************************/ 
void FEMBED_F77(int* decomp, int* vp, int* startPosn, int* endPosn) {
  DECOMP *decomp1;
  VPROC *vproc;
  
  decomp1 = lookupdecomp(*decomp);
  vproc = lookupvproc(*vp);
  
  FFlag = 1;
  embed(decomp1, vproc, *startPosn-1, *endPosn-1);
  FFlag = 0;
}

/***************************************************************
 * Parameters :
 *   decomp   : integer describing pointer to decomposition
 *   distInfo : Array describing distribution for each dimension
 *              (0) : Undistributed
 *              (1) : Block
 *              (2) : Cyclic ( Not Implemented )
 * Distributes a decomposition
 *****************************************************************/
void FDISTRIBUTE_F77(int* decomp, int* distInfo) {
  DECOMP *decomp1;
  char dists[10];
  int i;
  
  decomp1 = lookupdecomp(*decomp);
  
  for (i = 0; i < decomp1->rank; i++) {
    switch (distInfo[i]) {
    case 0: 
      dists[i] = '*';
      break;
    case 1: 
      dists[i] = 'B';
      break;
    case 2: 
      dists[i] = 'C';
      break;
    }
  }
  dists[i] = '\0';
  
  FFlag = 1;
  distribute(decomp1, dists);
  FFlag = 0;  
}

int IFSECTION_F77(int* decomp, int* dim, int* cnt, int** idx) {
  DECOMP *decomp1;
  int i, j, d[MAX_DIM], *x[MAX_DIM];

  decomp1 = lookupdecomp(*decomp);

  for (i = 0; i < decomp1->rank; ++i) {
    d[i] = dim[i] - 1;
    x[i]  = (int*)calloc(cnt[i], sizeof(int));
    for (j = 0; j < cnt[i]; ++j) {
      x[i][j] = idx[j][i] - 1;
    }  
  }

  FFlag = 1;
  i = section(decomp1, d, cnt, x);
  FFlag = 0;

  for (i = 0; i < decomp1->rank; ++i) {
    free(x[i]);
  }

  return i;
}

int IFPARTITION_F77(int* decomp, int* dim, int* blk, int* cnt, int* idx) {
  DECOMP *decomp1;
  int i, *b, *x;

  decomp1 = lookupdecomp(*decomp);
  
  b = (int*)calloc(*cnt, sizeof(int));
  x = (int*)calloc(*cnt, sizeof(int));
  for (i = 0; i < *cnt; ++i) {
    b[i] = blk[i] - 1;
    x[i] = idx[i] - 1;
  }

  FFlag = 1;
  i = partition(decomp1, *dim-1, b, *cnt, x);
  FFlag = 0;

  free(x);
  free(b);

  return i;
}

/*****************************************************************
 * Parameters :                                                  *
 *   vMach : Decomposition                                       *
 *   ndim  : Number of dimensions in Fortran array               *
 *   info0 : darray dimensions ( Not a parameter )               *
 *   info1 : dimension sizes                                     *
 *   info2 : No. of internal GhostCells                          *
 *   ext_gCells_l - number of external ghost cells on left in each dim
 *   ext_gCells_r - number of external ghost cells on right in each dim
 *   extra_flag : Where to put extra cells if lenght of dim is not    *
 *           exactly divisible by number of processors           *
 *           0. Default, which is option 4                       *
 *           1. on the left most processor                       *
 *           2. on the right most processor                      *
 *           3. Split equally if odd the extra one on left       *
 *           4. Split equally if odd the extra one on right      *
 *   info3 : decomposition dimensions to which darray            *
 *           dimension is aligned                                *
 * Returns : an index into the table holding the darray          *
 *           descriptors                                         *
 * Aligns the given array to given decomposition                 *
 *****************************************************************/
/* Modified by Serge Koren for merge with P++ */
int IFALIGN_F77(int* vMach, int* ndim, int* info1, int* left_gcells, int* right_gcells,
                 int* ext_gCells_l, int* ext_gCells_r, int* extra_flag, int* info3) {
  DECOMP *decomp;
  DARRAY *dArray;
  int i, info0[7], info4[7];
  
  decomp = lookupdecomp(*vMach);
  
  /* Info0 contains darray dimension numbers */
  /* Info3 contains decomposition dimension numbers, and -1 means the */
  /* darray dimension is not aligned to any decomp dimension */
  /* since we need to change the dimension numbers, copy them into info4 */
  /* and change the new array */
  for (i = 0; i < *ndim; i++) {
    info0[i] = i;
    info4[i] = info3[i];
    if (info4[i] > 0)
      info4[i]--;
  }
  
  FFlag = 1;
  dArray = align(decomp, *ndim, info0, info1, left_gcells, right_gcells, 
		 ext_gCells_l, ext_gCells_r, extra_flag, info4);
  FFlag = 0;

  return insert_da_table(dArray);
}

/******************************************************************
 * Parameters :
 *   ptr      : Starting address of the darray
 *   dim      : Dimension in question
 *   start    : Start index in the dimension
 *              ( Fortran Indexing starts at 1 )
 *   stride   : distance between elements
 *            : positive - increasing indexes
 *            : negative - decreasing indexes
 * Returns    : The local start index
 *******************************************************************/
int IFLALBND_F77(int* ptr, int* dim, int* start, int* stride) {
  int index;
  DARRAY *dArrayPtr;
  
  dArrayPtr = lookupda(*ptr);
  FFlag = 1;
  index = lalbnd(dArrayPtr, *dim-1, *start-1, *stride);
  FFlag = 0;
  
  return (index+1);
}

/******************************************************************
 * Parameters :
 *   ptr      : Starting address of the darray
 *   dim      : Dimension in question
 *   stop     : End index in the dimension
 *              ( Fortran Indexing starts at 1 )
 *   stride   : distance between elements
 *            : positive - increasing indexes
 *            : negative - decreasing indexes
 * Returns    : The local end index
 *******************************************************************/
int IFLAUBND_F77(int* ptr, int* dim, int* stop, int* stride) {
  int index;
  DARRAY *dArrayPtr;
  
  dArrayPtr = lookupda(*ptr);
  FFlag = 1;
  index = laubnd(dArrayPtr, *dim-1, *stop-1, *stride);
  FFlag = 0;
  
  return (index+1);
}

/*****************************************************************
 * Parameters :
 *   ptr      : Reference to the Darray descriptor
 *   ndims    : number of Dimensions in Darray
 *   fillVec  : offset vector for ghost cells
 *              (+) At the High End
 *              (-) At the Low end.
 * Returns    : An Integer describing pointer to schedule
 * Creates Schedule to Fill in Ghost Cells.
 *****************************************************************/
int IFGHOSTFILL_SCHED_F77(int* ptr, int* ndims, int* fillVec) {
  SCHED *schedPtr;
  DARRAY *dArrayPtr;
  
  dArrayPtr = lookupda(*ptr);
  
  FFlag = 1;
  schedPtr = ghostFillSched(dArrayPtr, *ndims, fillVec);
  FFlag = 0;
  
  return insert_sch_table(schedPtr);
}

/*****************************************************************
 * Parameters :
 *   ptr      : Reference to the Darray descriptor
 *   ndims    : number of Dimensions in Darray
 *   fillVec  : offset vector for ghost cells
 *              (+) At the High End
 *              (-) At the Low end.
 * Returns    : An Integer describing pointer to schedule
 * Creates Schedule to Fill in all Ghost Cells related to the fill vector
 *****************************************************************/
int IFGHOSTFILLSPAN_SCHED_F77(int* ptr, int* ndims, int* fillVec) {
   SCHED *schedPtr;
   DARRAY *dArrayPtr;
   
   dArrayPtr = lookupda(*ptr);
   
   FFlag = 1;
   schedPtr = ghostFillSpanSched(dArrayPtr, *ndims, fillVec);
   FFlag = 0;
   
   return insert_sch_table(schedPtr);
}

/*****************************************************************
 * Parameters :
 *   ptr      : Reference to the Darray descriptor
 *
 * Returns    : An Integer describing pointer to schedule
 * Creates Schedule to Fill in all Ghost Cells for the darray pointed to by ptr
 *****************************************************************/
int IFGHOSTFILLALL_SCHED_F77(int* ptr) {
  SCHED *schedPtr;
  DARRAY *dArrayPtr;
  
  dArrayPtr = lookupda(*ptr);
  
  FFlag = 1;
  schedPtr = ghostFillAllSched(dArrayPtr);
  FFlag = 0;
  
  return insert_sch_table(schedPtr);
}

/*****************************************************************
 * Parameters :
 *   ptr      : Reference to the Darray descriptor
 *   dim      : Fill Dimension
 *   amtFill  : Amount of Fill in.
 *              (+) At the High End
 *              (-) At the Low end.
 * Returns    : An Integer describing pointer to schedule
 * Creates Schedule to Fill in Ghost Cells.
 *****************************************************************/
int IFEXCH_SCHED_F77(int* ptr, int* dim, int* amtFill) {
  SCHED *schedPtr, *exchSched();
  DARRAY *dArrayPtr;
  
  dArrayPtr = lookupda(*ptr);
  
  FFlag = 1;
  schedPtr = exchSched(dArrayPtr, *dim-1, *amtFill);
  FFlag = 0;
  
  return insert_sch_table(schedPtr);
}

/*******************************************************************/
/* Parameters :                                                    */
/*   scrPtr   : Src dArray descriptor                              */
/*   destPtr  : Dest dArray descriptor                             */
/*   ndim     : Number of dimensions                               */
/*   s_dim    :    Source Dimensions				   */
/*   s_lo     :    Source Low Indices				   */
/*   s_hi     :    Source High Indices				   */
/*   s_stride :    Source Strides				   */
/*   d_dim    :    Dest Dimensions				   */
/*   d_lo     :    Dest Low Indices				   */
/*   d_hi     :    Dest High Indices				   */
/*   d_stride :    Dest Strides					   */
/*   Returns : an integer describing pointer to schedule	   */ 
/*   Function : Constructs a schedule describing data motion for   */
/*		copying a subarray from source into a subarray of  */
/*		destination					   */
/*******************************************************************/
int IFSUBARRAY_SCHED_F77(int* srcPtr, int* destPtr, int* ndim, 
		      int* s_dim, int* s_lo, int* s_hi, int* s_stride,
		      int* d_dim, int* d_lo, int* d_hi, int* d_stride) {
  SCHED *schedPtr;
  DARRAY *srcDAPtr, *destDAPtr;
  int srcDims[7], destDims[7];
  int sLos[7], dLos[7];
  int sHis[7], dHis[7];
  int i;
  
  FFlag = 1;
  
  srcDAPtr  = lookupda(*srcPtr);
  destDAPtr = lookupda(*destPtr);
  
  for (i = 0; i < *ndim; i++) {
    srcDims[i] = s_dim[i]-1;
    destDims[i] = d_dim[i]-1;
    sLos[i] = s_lo[i]-1;
    dLos[i] = d_lo[i]-1;
    sHis[i] = s_hi[i]-1;
    dHis[i] = d_hi[i]-1;
  }
  
  schedPtr = subArraySched(srcDAPtr, destDAPtr, *ndim,
			   srcDims,  sLos, sHis, s_stride,
			   destDims, dLos, dHis, d_stride);
  FFlag = 0;
  
  return insert_sch_table(schedPtr); 
}

/*********************************************************************
 * The data move functions move data for various data types
 * Parameters :
 *   src      : Starting address of Source Darray
 *   sched    : schedule describing data movement
 *   dest     : Starting address of Destination Darray
 *********************************************************************/
/* Integer data move */
void FIDATA_MOVE_F77(int* src, int* sched, int* dest) {
  SCHED *schedPtr;
  
  schedPtr = lookupsch(*sched);
  FFlag = 1;
  iDataMove((int*)src, schedPtr, (int*)dest);
  FFlag = 0;
}

/* Single precision float data move */
void FFDATA_MOVE_F77(int* src, int* sched, int* dest) {
  SCHED *schedPtr;
  
  schedPtr = lookupsch(*sched);
  FFlag = 1;
  fDataMove((float*)src, schedPtr, (float*)dest);
  FFlag = 0;
}

/* Double precision float data move */
void FDDATA_MOVE_F77(int* src, int* sched, int* dest) {
  SCHED *schedPtr;
  
  schedPtr = lookupsch(*sched);
  FFlag = 1;
  dDataMove((double*)src, schedPtr, (double*)dest);
  FFlag = 0;
}

/* Character data move */
void FCDATA_MOVE_F77(int* src, int* sched, int* dest) {
  SCHED *schedPtr;
  
  schedPtr = lookupsch(*sched);
  FFlag = 1;
  cDataMove((char*)src, schedPtr, (char*)dest);
  FFlag = 0;
}

/*****************************************************************
 * returns in info the size in each dimension of the local array
 * This routine is used for memory allocation
 ******************************************************************/
void FLASIZES_F77(int* ptr, int* info) {
  DARRAY *dArray;  
  dArray = lookupda(*ptr);
  FFlag = 1;
  laSizes(dArray, info);
  FFlag = 0;
}

/*********************************
 *
 * Frees the schedule
 *
 **********************************/
void FFREE_SCHED_F77(int* sched) {
  SCHED *schedPtr;  
  schedPtr = lookupsch(*sched);
  parti_free_sched(schedPtr);
}

/****************************************************************
 *
 *  remove all stored exchange schedules
 *
 *****************************************************************/
void REMOVE_EXCH_SCHEDS_F77() {
  remove_exch_scheds();
}

/****************************************************************
 *
 *  remove all stored subarray_exch schedules
 *
 *****************************************************************/
void REMOVE_SUBARRAY_SCHEDS_F77() {
  remove_subarray_scheds();
}

/*****************************************************************
 * for the array pointed to by ptr determine in GLOBAL terms the
 * index of the first value in dimension "dim" stored in the
 * current processor
 ******************************************************************/
int IFGLBND_F77(int* ptr, int* dim) {
  DARRAY *dArray;
  int arrayDim;  
  dArray = lookupda(*ptr);
  arrayDim  = *dim - 1;
  return gLBnd(dArray, arrayDim);
}

/*****************************************************************
 * for the array pointed to by ptr determine in GLOBAL terms the
 * index of the last value in dimension "dim" stored in the
 * current processor
 ******************************************************************/
int IFGUBND_F77(int* ptr, int* dim) {
  DARRAY *dArray;
  int arrayDim;
  dArray = lookupda(*ptr);  
  arrayDim  = *dim - 1;
  return gUBnd(dArray, arrayDim);
}

/*****************************************************************
 * for the array pointed to by ptr and the GLOBAL index "index" in
 * dimension "dim", returns the corresponding LOCAL index value
 *
 ******************************************************************/
int IFGLOBAL_TO_LOCAL_F77(int* ptr, int* index, int* dim) {
  DARRAY *dArray;
  int arrayDim, gindex;
  dArray = lookupda(*ptr);
  gindex = *index - 1;
  arrayDim  = *dim - 1;
  return globalToLocal(dArray, gindex, arrayDim);
}

/*****************************************************************
 * for the array pointed to by ptr and the GLOBAL index "index" in
 * dimension "dim", returns the corresponding LOCAL index value,
 * including internal ghost cells
 *
 ******************************************************************/
int IFGLOBAL_TO_LOCAL_WITH_GHOST_F77(int* ptr, int* index, int* dim) {
  DARRAY *dArray;
  int arrayDim, gindex;
  dArray = lookupda(*ptr);  
  gindex = *index - 1;
  arrayDim  = *dim - 1;
  return globalToLocalWithGhost(dArray, gindex, arrayDim);
}

/*****************************************************************
 * for the array pointed to by ptr and the LOCAL index "index" in
 * dimension "dim", returns the corresponding GLOBAL "index" value
 *
 ******************************************************************/
int IFLOCAL_TO_GLOBAL_F77(int* ptr, int* index, int* dim) {
  DARRAY *dArray;
  int arrayDim;
  int lindex = *index;
  dArray = lookupda(*ptr);  
  arrayDim  = *dim - 1;
  return localToGlobal(dArray, lindex, arrayDim);
}

/*****************************************************************
 * for the array pointed to by ptr and the LOCAL index "index" in
 * dimension "dim", returns the corresponding GLOBAL "index" value,
 * including internal ghost cells
 *
 ******************************************************************/
int IFLOCAL_TO_GLOBAL_WITH_GHOST_F77(int* ptr, int* index, int* dim) {
  DARRAY *dArray;
  int arrayDim;
  int lindex = *index;
  /* int num_ghost; */
  int return_val;
  
  dArray = lookupda(*ptr);
  arrayDim  = *dim - 1;
  return localToGlobalWithGhost(dArray, lindex, arrayDim);
}

/* NOT PART OF PUBLISHED INTERFACE */
void FCLEANUP_F77() {
  destroy_sch_table();
  destroy_da_table();
  destroy_decomp_table();
  destroy_vproc_table();
}

#endif /* FORTRAN */
