/* 
 * Auth: Alan Sussman (als@cs.umd.edu)
 *       University of Maryland, College Park
 * Desc: 
 */

#include "hash.h"
#include "addition.h"
#include "list.h"
#include "utils.h"
#include <stdlib.h>
#include <assert.h>
#include <stdio.h>

#define SCH_TABLE_SIZE 1024
static int sch_done_init = 0;

static SCHED *sch_table[SCH_TABLE_SIZE];
static int sch_last = 0;  /* points to first open position in sch_table */

#define DA_TABLE_SIZE 256
static int da_done_init = 0;

static DARRAY *da_table[DA_TABLE_SIZE];
static int da_last = 0;  /* points to first open position in da_table */

#define DECOMP_TABLE_SIZE 256
static int decomp_done_init = 0;

static DECOMP *decomp_table[DECOMP_TABLE_SIZE];
static int decomp_last = 0; /* points to first open position in decomp_table */

#define VPROC_TABLE_SIZE 64
static int vproc_done_init = 0;

static VPROC *vproc_table[VPROC_TABLE_SIZE];
static int vproc_last = 0; /* points to first open position in vproc_table */

#define SCHED_TABLE_SIZE 256

static list subarray_table[SCHED_TABLE_SIZE]; /* for subarray schedules */
static int sub_done_init = 0;

static list exch_table[SCHED_TABLE_SIZE];     /* for exchange schedules */
static int exch_done_init = 0;

typedef struct { 
  SCHED* subarray_sched;
  DARRAY* srcDA;
  DARRAY* destDA;
  int numDims;
  int srcDims[MAX_DIM], sLos[MAX_DIM], sHis[MAX_DIM], sStrides[MAX_DIM];
  int destDims[MAX_DIM], dLos[MAX_DIM], dHis[MAX_DIM], dStrides[MAX_DIM];
} sub_cell;

typedef struct { 
  SCHED* exch_sched;
  DARRAY* dArrayPtr;
  int ndims;
  int fillVec[MAX_DIM];
} exch_cell;

void init_sch_table() {
  int i;
  
  if (sch_done_init)
    return;  

  for (i = 0; i < SCH_TABLE_SIZE; i++) {
    sch_table[i] = NULL;
  }

  sch_done_init = 1;
}

void destroy_sch_table() {
  int i;

  for (i = 0; i < sch_last; i++) {
    if (sch_table[i]) delete_SCHED(sch_table[i]);
  }
  sch_last = 0;

  sch_done_init = 0;
}

/* NO DUPLICATES */
int insert_sch_table(SCHED* sched) {
  int i, empty = -1;
  
  if (!sch_done_init)
    init_sch_table();
  
  if (sched == NULL) 
    return -1; /* the NULL schedule index */
  
  /* first check to see that it's not already in the table */
  for (i = 0; i < sch_last; i++) {
    if (sch_table[i] == sched) {
      return i;
    }
    if (empty < 0 && sch_table[i] == NULL) {
      /* first empty entry */
      empty = i;
    }
  }
  
  /* insert into the empty entry */
  if (empty >= 0) {
    sch_table[empty] = sched;
    sched->referenceCount++;
    return empty;
  }
  
  if (sch_last >= SCH_TABLE_SIZE) {
    fprintf(stderr, "No more room in schedule table");
    return -1;
  }
  
  sch_table[sch_last] = sched;
  
  return sch_last++;
}

void remove_sch_table(SCHED* sched) {
  int i;
  
  for (i = 0; i < sch_last; i++) {
    if (sch_table[i] == sched) {
      sch_table[i] = NULL;
      return;
    }
  }
}

SCHED* lookupsch(int x) {
  if (!sch_done_init)
    init_sch_table();
  
  if (x == -1) 
    return NULL;  /* the NULL schedule index */
  
  if (x >= sch_last || x < 0) {
    fprintf(stderr, "Tried to index a non-existent entry in the schedule table");
    return 0;
  }

  return (sch_table[x]);
}

void init_da_table() {
  int i;
  
  if (da_done_init) 
    return;
  
  for (i = 0; i < DA_TABLE_SIZE; i++) {
    da_table[i] = NULL;
  }

  da_done_init = 1;
}

void destroy_da_table() {
  int i;
  
  for (i = 0; i < da_last; i++) {
    if (da_table[i]) 
      delete_DARRAY(da_table[i]);
  }
  da_last = 0;

  da_done_init = 0;
}

int insert_da_table(DARRAY* da) {
  if (!da_done_init)
    init_da_table();
  
  if (da_last >= DA_TABLE_SIZE) {
    fprintf(stderr, "No more room in darray table");
    return -1;
  }
  
  da_table[da_last] = da;
  da->referenceCount++;
  
  return da_last++;
}

DARRAY* lookupda(int x) {
  if (!da_done_init)
    init_da_table();
  
  if (x >= da_last || x < 0) {
    fprintf(stderr, "Tried to index a non-existent descriptor in the darray table");
    return 0;
  }

  return da_table[x];
}

void init_decomp_table() {
  int i;
  
  if (decomp_done_init) 
    return;
  
  for (i = 0; i < DECOMP_TABLE_SIZE; i++) {
    decomp_table[i] = NULL;
  }

  decomp_done_init = 1;
}

void destroy_decomp_table() {
  int i;

  for (i = 0; i < decomp_last; i++) {
    if (decomp_table[i]) delete_DECOMP(decomp_table[i]);
  }
  decomp_last = 0;

  decomp_done_init = 0;
}

int insert_decomp_table(DECOMP* decomp) {
  if (!decomp_done_init)
    init_decomp_table();
  
  if (decomp_last >= DECOMP_TABLE_SIZE) {
    fprintf(stderr, "No more room in decomposition table");
    return -1;
  }
  
  decomp_table[decomp_last] = decomp;
  
  return decomp_last++;
}

DECOMP* lookupdecomp(int x) {
  if (!decomp_done_init)
    init_decomp_table();
  
  if (x >= decomp_last || x < 0) {
    fprintf(stderr, "Tried to index a non-existent entry in the decomposition table");
    return 0;
  }
  return decomp_table[x];
}

void init_vproc_table() {
  int i;
  
  if (vproc_done_init) 
    return;
  
  for (i = 0; i < VPROC_TABLE_SIZE; i++) {
    vproc_table[i] = NULL;
  }

  vproc_done_init = 1;
}

void destroy_vproc_table() {
  int i;

  for (i = 0; i < vproc_last; i++) {
    if (vproc_table[i]) delete_VPROC(vproc_table[i]);
  }
  vproc_last = 0;

  vproc_done_init = 0;
}

int insert_vproc_table(VPROC* vproc) {
  if (!vproc_done_init)
    init_vproc_table();
  
  if (vproc_last >= VPROC_TABLE_SIZE) {
    fprintf(stderr, "No more room in vproc table");
    return -1;
  }
  
  vproc_table[vproc_last] = vproc;
  
  return vproc_last++;
}

VPROC* lookupvproc(int x) {
  if (!vproc_done_init)
    init_vproc_table();
  
  if (x >= vproc_last || x < 0) {
    fprintf(stderr, "Tried to index a non-existent entry in the vproc table");
    return 0;
  }

  return (vproc_table[x]);
}

void init_exch_table() {
  int i;
  
  for (i = 0; i < SCHED_TABLE_SIZE; i++) {
    exch_table[i] = create_list();
  }

  exch_done_init = 1;
}

void destroy_exch_table() {
  int i;
  exch_cell* cell;
  
  if (!exch_done_init)
    return;
  
  for (i = 0; i < SCHED_TABLE_SIZE; i++) {
    FOREACH(cell, exch_table[i]) {
      delete_SCHED(cell->exch_sched);
      cell->exch_sched = NULL;
      pfree("exch_cell", sizeof(exch_cell), cell);
    }
    destroy_list(exch_table[i]);
  }

  exch_done_init = 0;
}

int exch_hash(DARRAY* darray, int ndims, int* fillVec) {
  int i, hash = 0;
  char* p = (char*)darray;
  
  hash = ((p[0] ^ p[1] ^ p[2] ^ p[3]) + ndims);
  for (i = 0; i < ndims; i++) {
    hash += fillVec[i];
  }
  hash &= 0xff;

#if PARTI_DEBUG > 2
  printf("Node %d: generating exchange hash value %d\n", PARTI_myproc(), hash);
#endif
  return hash;
}

void insert_exch_table(SCHED* sched, DARRAY* darray, int ndims, int* fillVec) {
  int i, j;
  exch_cell* cell;
  
  if (!exch_done_init)
    init_exch_table();

  i = exch_hash(darray, ndims, fillVec);
  sched->hash = i;
  cell = (exch_cell*)pmalloc("exch_cell", sizeof(exch_cell));
  cell->exch_sched = sched;
  sched->referenceCount++;
  cell->dArrayPtr = darray;
  cell->ndims = ndims;
  for (j = 0; j < ndims; j++) {
    cell->fillVec[j] = fillVec[j];
  }

  insert_list(cell, exch_table[i]);
}

void delete_exch_table(SCHED* sched) {
  int i = sched->hash;
  exch_cell* cell;
  
  if (!exch_done_init)
    return;
  
  FOREACH(cell, exch_table[i]) {
    if (cell->exch_sched == sched) {
      cell = remove_list(cell, exch_table[i]);
      delete_SCHED(cell->exch_sched);
      cell->exch_sched = NULL;
      pfree("exch_cell", sizeof(exch_cell), cell);
      break;
    }
  }
}

void clear_exch_table(DARRAY* darray) {
  exch_cell* cell;
  int i, count = 0;

  if (!exch_done_init)
    return;
  
  for (i = 0; i < SCHED_TABLE_SIZE; i++) {
    exch_table[i]->fptr = FIRST(exch_table[i]);
    cell = exch_table[i]->fptr->item;
    while (STILL_IN(exch_table[i]->fptr, exch_table[i])) {
      exch_table[i]->fptr = NEXT(exch_table[i]->fptr);
      if (cell->dArrayPtr == darray) {
	cell = remove_list(cell, exch_table[i]);
	delete_SCHED(cell->exch_sched);
	cell->exch_sched = NULL;
	pfree("exch_cell", sizeof(exch_cell), cell);
      }
      cell = exch_table[i]->fptr->item;
    }
  }
  
#if PARTI_DEBUG > 2
  for (i = 0; i < SCHED_TABLE_SIZE; i++) {
    count += length_list(exch_table[i]);
  }
  printf("exch_table_size=%d\n", count);
#endif
}

void verify_exch_table() {
  exch_cell* cell;
  int i, j;
  
  for (i = 0; i < SCHED_TABLE_SIZE; ++i) {
    FOREACH (cell, exch_table[i]) {
      for (j = 0; j < PARTI_numprocs(); ++j) {
	if (cell->exch_sched->rMsgSz[j] > 0) 
	  if (cell->exch_sched->rData[j]->proc != j)
	    printf("sched id %d, recv proc %d != %d\n", cell->exch_sched->ID_Number, 
		   cell->exch_sched->rData[j]->proc, j);	
	if (cell->exch_sched->sMsgSz[j] > 0) 
	  if (cell->exch_sched->sData[j]->proc != j)
	    printf("sched id %d, send proc %d != %d\n", cell->exch_sched->ID_Number,
		   cell->exch_sched->sData[j]->proc, j);
      }          
    }
  }
}

SCHED* lookup_exch(DARRAY* darray, int ndims, int* fillVec) {
  int i, j, found;
  exch_cell* cell;
  
  if (!exch_done_init)
    init_exch_table();

  i = exch_hash(darray, ndims, fillVec);  
  FOREACH(cell, exch_table[i]) {
    found = 1;
    if (cell->dArrayPtr != darray || cell->ndims != ndims)
      found = 0;
    for (j = 0; j < ndims; j++) {
      if (cell->fillVec[j] != fillVec[j])
	found = 0;
    }
    
    if (found) {
      cell->exch_sched->referenceCount++;
      return cell->exch_sched;
    }
  }
  
  return NULL;
}

void init_sub_table() {
  int i;
  
  for (i = 0; i < SCHED_TABLE_SIZE; i++) {
    subarray_table[i] = create_list();
  }
  
  sub_done_init = 1;
}

void destroy_sub_table() {
  int i;
  sub_cell* cell;

  if (!sub_done_init)
    return;

  for (i = 0; i < SCHED_TABLE_SIZE; i++) {
    FOREACH(cell, subarray_table[i]) {
      delete_SCHED(cell->subarray_sched);
      cell->subarray_sched = NULL;
      pfree("sub_cell", sizeof(sub_cell), cell);
    }
    destroy_list(subarray_table[i]);
  }
  
  sub_done_init = 0;
}

int sub_hash(DARRAY* srcDA, DARRAY* destDA, int numDims,
	     int* srcDims, int* sLos, int* sHis, int* sStrides,
	     int* destDims, int* dLos, int* dHis, int* dStrides) {
  int hash, i;
  char* p1 = (char*)srcDA;
  char* p2 = (char*)destDA;
  
  hash = (p1[0] ^ p1[1] ^ p1[2] ^ p1[3] ^ p2[0] ^ p2[1] ^ p2[2] ^ p2[3]);
  hash += numDims;
  for (i = 0; i < numDims; i++) {
    hash += srcDims[i] + sLos[i] + sHis[i] + sStrides[i];
    hash += destDims[i] + dLos[i] + dHis[i] + dStrides[i];
  }
  hash &= 0xff;

  /* printf("Node %d: generating subarray hash value %d\n", PARTI_myproc(), hash); */
  return hash;
}

void insert_sub_table(SCHED* sched, DARRAY* srcDA, DARRAY* destDA, int numDims,
		      int* srcDims, int* sLos, int* sHis, int* sStrides,
		      int* destDims, int* dLos, int* dHis, int* dStrides) {
  int i, j;
  sub_cell* cell;

  if (!sub_done_init)
    init_sub_table();

  i = sub_hash(srcDA, destDA, numDims,
	       srcDims, sLos, sHis, sStrides,
	       destDims, dLos, dHis, dStrides);
  sched->hash = i;  
  cell = (sub_cell*)pmalloc("sub_cell", sizeof(sub_cell));
  cell->subarray_sched = sched;
  sched->referenceCount++;
  cell->srcDA = srcDA;
  cell->destDA = destDA;
  cell->numDims = numDims;
  for (j = 0; j < numDims; j++) {
    cell->srcDims[j] = srcDims[j];
    cell->sLos[j] = sLos[j];
    cell->sHis[j] = sHis[j];
    cell->sStrides[j] = sStrides[j];
    cell->destDims[j] = destDims[j];
    cell->dLos[j] = dLos[j];
    cell->dHis[j] = dHis[j];
    cell->dStrides[j] = dStrides[j];
  }

  /* Initialize the rest to avoid purify URM errors */
  for (j = numDims; j < MAX_DIM; j++) {
    cell->srcDims[j] = 0;
    cell->sLos[j] = 0;
    cell->sHis[j] = 0;
    cell->sStrides[j] = 0;
    cell->destDims[j] = 0;
    cell->dLos[j] = 0;
    cell->dHis[j] = 0;
    cell->dStrides[j] = 0;
  }
	       
  insert_list(cell, subarray_table[i]);
}

void delete_sub_table(SCHED* sched) {
  int i = sched->hash;
  sub_cell* cell;

  if (!sub_done_init)
    return;
  
  FOREACH(cell, subarray_table[i]) {
    if (cell->subarray_sched == sched) {
      cell = remove_list(cell, subarray_table[i]);
      delete_SCHED(cell->subarray_sched);
      cell->subarray_sched = NULL;
      pfree("sub_cell", sizeof(sub_cell), cell);
      break;
    }
  }
}

void clear_sub_table(DARRAY* darray) {
  sub_cell* cell;
  int i;

  if (!sub_done_init)
    return;
  
  for (i = 0; i < SCHED_TABLE_SIZE; i++) {
    subarray_table[i]->fptr = FIRST(subarray_table[i]);
    cell = subarray_table[i]->fptr->item;
    while (STILL_IN(subarray_table[i]->fptr, subarray_table[i])) {
      subarray_table[i]->fptr = NEXT(subarray_table[i]->fptr);
      if (cell->srcDA == darray) {
	cell = remove_list(cell, subarray_table[i]);
	delete_SCHED(cell->subarray_sched);
	cell->subarray_sched = NULL;
	pfree("sub_cell", sizeof(sub_cell), cell);
      }
      cell = subarray_table[i]->fptr->item;
    }
  }
}

SCHED* lookup_sub(DARRAY* srcDA, DARRAY* destDA, int numDims,
		  int* srcDims, int* sLos, int* sHis, int* sStrides,
		  int* destDims, int* dLos, int* dHis, int* dStrides) {
  int i, found, j;
  sub_cell* cell;

  if (!sub_done_init)
    init_sub_table();

  i = sub_hash(srcDA, destDA, numDims,
	       srcDims, sLos, sHis, sStrides,
	       destDims, dLos, dHis, dStrides);
  FOREACH(cell, subarray_table[i]) {
    found = 1;
    if (cell->srcDA != srcDA || cell->destDA != destDA || cell->numDims != numDims)
      found = 0;
    for (j = 0; j < numDims; j++) {
      if (cell->srcDims[j] != srcDims[j] || cell->sLos[j] != sLos[j] ||
	  cell->sHis[j] != sHis[j] || cell->sStrides[j] != sStrides[j])
	found = 0;
      if (cell->destDims[j] != destDims[j] || cell->dLos[j] != dLos[j] ||
	  cell->dHis[j] != dHis[j] || cell->dStrides[j] != dStrides[j])
	found = 0;
    }    
    if (found) {
      cell->subarray_sched->referenceCount++;
      return cell->subarray_sched;
    }
  }
  
  return NULL;
}

void cleanup_da_table() {
  int i = 0;
#if PARTI_DEBUG > 2
  printf("Inside of cleanup_da_table\n");
#endif
  
  for (i = 0; i < DA_TABLE_SIZE; i++) {
    if (da_table[i] != NULL) {
      delete_DARRAY(da_table[i]);
      da_table[i] = NULL;
    }
  }
}

void cleanup_subarray_table() {
#if PARTI_DEBUG > 2
  printf("Inside of cleanup_subarray_table\n");
#endif
  destroy_sub_table();
}

void cleanup_exch_table() {
#if PARTI_DEBUG > 2
  printf("Inside of cleanup_exch_table\n");
#endif
  destroy_exch_table();
}
