/*
 * Auth: Alan Sussman (als@cs.umd.edu)
 *       University of Maryland, College Park
 * Desc:
 */

/*
 * This file implements a doubly linked, circular lists of things
 */  
#include "list.h"
#include <assert.h>
#include <stdlib.h>

/* create an empty list */
list create_list() {
  list l;
  cell* c;
  
  /* empty lists always have at least one thing in them */
  l = (list)malloc(sizeof(struct _list));
  l->referenceCount = 0;
  c = (cell*)malloc(sizeof(cell));
  c->referenceCount = 0;
  l->dummy = c;
  c->item = NULL;
  c->next = c;
  c->prev = c;

  return l;
}

/* destroy a (possibly nonempty) list */
void destroy_list(list l) {
  cell* c;
  cell* t;
  
  c = FIRST(l);
  while (STILL_IN(c, l)) {
    t = NEXT(c);    
    assert(c->referenceCount >= 0);
    if (c->referenceCount-- == 0) {
      free(c);
    }
    c = t;
  }
  
  assert(l->dummy->referenceCount >= 0);
  if (l->dummy->referenceCount-- == 0) {
    free(l->dummy);
  }

  assert(l->referenceCount >= 0);
  if (l->referenceCount-- == 0) {
    free(l);
  }
  
  l = NULL;
}
      
/* insert b onto the end of list */
void insert_list(void* b, list l) {
  cell* c;

  c = (cell*)malloc(sizeof(cell));
  c->referenceCount = 0;
  c->item = b;
  c->next = l->dummy;
  c->prev = l->dummy->prev;
  l->dummy->prev->next = c;
  l->dummy->prev = c;
}

/*
  remove a cell from a list
  if x is NIL, then remove the first cell
  otherwise, find x and remove it.
  if x is not in the list, return NIL
*/  
void* remove_list(gtype* x, list l) {
  cell* c;
  void* b;
  
  assert(!EMPTY(l));
  if (x) { /* find x and remove it */
    for (c = FIRST(l); STILL_IN(c, l); c = NEXT(c)) {
      if (c->item == x)
        break;
    }

    if (STILL_IN(c, l)) { /* we found x */
      c->prev->next = c->next;
      c->next->prev = c->prev;
      assert(c->referenceCount >= 0);
      if (c->referenceCount-- == 0) {
	free(c);
      }
      return x;

    } else { /* x is not in the list */
      return NULL;
    }

  } else { /* remove the first element */
    c = FIRST(l);
    FIRST(l) = c->next;
    (FIRST(l))->prev = l->dummy;
    b = c->item;
    assert(c->referenceCount >= 0);
    if (c->referenceCount-- == 0) {
      free(c);
    }
    return b;
  }
}

/*
  remove every item in l1 from l2
*/
void remove_list_list(list l1, list l2) {
  cell* c;
  
  for (c = FIRST(l1); STILL_IN(c, l1); c = NEXT(c)) {
    if (member_list(c->item, l2)) {
      remove_list(c->item, l2);
    }
  }
}

gtype* member_list(void* b, list l) {
  cell* x;
  
  for (x = FIRST(l); STILL_IN(x, l); x = NEXT(x)) {
    if (x->item == b) {
      return x->item;
    }
  }

  return NULL;
}

gtype* eq_member_list(void* b, list l, pfunc f) {
  cell* x;
  
  for (x = FIRST(l); STILL_IN(x, l); x = NEXT(x)) {
    if (f(x->item, b)) {
      return x->item;
    }
  }

  return NULL;
}

int length_list(list l) {
  cell* x;
  int count = 0;
  
  for (x = FIRST(l); STILL_IN(x, l); x = NEXT(x)) {
    count++;
  }

  return count;
}

/*
  Duplicate a list
*/
list dup_list(list l) {
  list l2 = create_list();
  cell* c;
  
  for (c = FIRST(l); STILL_IN(c, l); c = NEXT(c)) {
    insert_list(c->item, l2);
  }

  return l2;
}

/*
  Append two lists together.
  l1 <-- l1 + l2
  ***** NOTE THIS DESTROYS l2!!!!! *****
  if you don't want l2 destroyed, you'd better copy it first
*/
void append_list(list l1, list l2) {
  l1->dummy->prev->next = l2->dummy->next; /* end of l1 points to l2 */
  l2->dummy->next->prev = l1->dummy->prev; /*      vice versa        */
  l2->dummy->prev->next = l1->dummy; /* end of the list is l1->dummy */
  l1->dummy->prev = l2->dummy->prev;
  
  assert(l2->dummy->referenceCount >= 0);
  if (l2->dummy->referenceCount-- == 0) {
    free(l2->dummy);
  }
  assert(l2->referenceCount >= 0);
  if (l2->referenceCount-- == 0) {
    free(l2);
  }
}

/*
  compute the intersection of two lists
*/
list intersection_list(list l1, list l2) {
  list l = create_list();
  cell* c;
  
  for (c = FIRST(l1); STILL_IN(c, l1); c = NEXT(c)) {
    if (member_list(c->item, l2)) {
      insert_list(c->item, l);
    }
  }
  
  return l;
}

/*
  compute the union of two lists
*/
list union_list(list l1, list l2) {
  list l = dup_list(l1);
  cell* c;
  
  for (c = FIRST(l2); STILL_IN(c, l2); c = NEXT(c)) {
    if (!member_list(c->item, l)) {
      insert_list(c->item, l);
    }
  }
  return l;
}

/*
  replace one item in a list with another item
  (this function is useful in ordered lists)
*/
void replace_list(gtype* i1, gtype* i2, list l) {
  cell* c;
  for (c = FIRST(l); STILL_IN(c, l); c = NEXT(c)) {
    if (c->item == i1) {
      c->item = i2;
      break;
    }
  }
}

void behead_list(list l) {
  remove_list((void*)0, l);
}
