/*
 * Auth: Alan Sussman (als@cs.umd.edu)
 *       University of Maryland, College Park
 * Desc:
 */

#include "addition.h"
#include "hash.h"
#include "port.h"
#include "config.h"
#include <stdlib.h>
#include <assert.h>

#ifdef USE_MPI
int PARTI_numberOfMessagesSent;
int PARTI_numberOfMessagesRecieved;
int PARTI_messagePassingInterpretationReport;

static MPI_Comm parti_mpi_comm;

int PARTI_MPI_setup(MPI_Comm* grp) {
  int return_status;
  assert(grp != NULL);
  return_status = MPI_Comm_dup(*grp, &parti_mpi_comm);
  assert(PARTI_MPI_numprocs() < MAX_NODES);
  return 0;
}

void PARTI_MPI_cleanup() {
  cleanup_after_PARTI(); /* destroy hash tables */
  MPI_Comm_free(&parti_mpi_comm);
}

/*
   We have included the PARTI_MPI_numprocs function so that something in this file
   could be compiled and so we could avoid a warning about "empty translation unit"
*/
int PARTI_MPI_numprocs() {
  /* Use a static variable to call the MPI function only once and hold it locally. */
  static int Size = 0;
  
  int flag;
  int return_status;
  
  /* D. Quinlan: Rewrote this function to avoid letting any MPI function from being called
     more than once (at the beginning, instead of at the end where it is an
     error since some internal data is removed in the cleanup process.  This error
     effects the use of MPI on the Blue Pacific machine only. */
  
  if (Size == 0) {
    /* finally we get the number of processors! */
    return_status = MPI_Comm_size(parti_mpi_comm, &Size);
    assert(Size > 0);
  }
  
  return Size;
}

int PARTI_MPI_myproc() {
  int Rank;
  int return_status;
  
  return_status = MPI_Comm_rank(parti_mpi_comm, &Rank);
  
  return Rank;
}

void PARTI_MPI_killall() {
  /*    MPI_Finalize(); */
  MPI_Abort(parti_mpi_comm, 99);
}

int PARTI_MPI_csend(int tag, char* buf, int len, int node) {
  int return_status;
  
  /* part of P++ diagnostics */
  PARTI_numberOfMessagesSent++;
  return_status = MPI_Send((void*)buf, len, MPI_CHAR, node, tag, parti_mpi_comm);
  
  return 0;
}

int PARTI_MPI_crecv(int tag, char* buf, int len, int node) {
  MPI_Status status;
  int return_status;
  
  /* part of P++ diagnostics */
  PARTI_numberOfMessagesRecieved++;
  return_status = MPI_Recv((void*)buf, len, MPI_CHAR, node, tag, parti_mpi_comm, &status);

  return 0;
}

int PARTI_MPI_isend(int tag, char* buf, int len, int node, MPI_Request* gid) {
  int return_status;
  
  /*printf ("Calling MPI_Isend() in PARTI on processor %d \n",PARTI_myproc());*/ 
  /* part of P++ diagnostics */
  PARTI_numberOfMessagesSent++;  
  assert(gid != NULL);
  assert(buf != NULL);
  return_status = MPI_Isend((void*)buf, len, MPI_CHAR, node, tag, parti_mpi_comm, gid);
  
  return 0;
}

int PARTI_MPI_irecv(int tag, char* buf, int len, int node, MPI_Request* gid) {
  int return_status;
  
  /*printf ("Calling MPI_Irecv() in PARTI on processor %d \n",PARTI_myproc());*/
  /* part of P++ diagnostics */
  PARTI_numberOfMessagesRecieved++;
  assert(gid != NULL);
  assert(buf != NULL);
  return_status = MPI_Irecv((void*)buf, len, MPI_CHAR, node, tag, parti_mpi_comm, gid);

  return 0;
}

int PARTI_MPI_msgwait(MPI_Request* gid) {
  MPI_Status status;
  int return_status;
  
  return_status = MPI_Wait(gid, &status);
  return 0;
}

int PARTI_MPI_msgdone(MPI_Request* gid) {
  MPI_Status status;
  int flag;
  int return_status; 
  
  return_status = MPI_Test(gid, &flag, &status);
  return flag;
}

int PARTI_MPI_gsync() {
  int return_status = MPI_Barrier(parti_mpi_comm);
  return return_status;
}

#ifdef FORTRAN

#define PARTISETUP_F77   F77_FUNC(partisetup, PARTISETUP)
#define PARTICLEANUP_F77 F77_FUNC(particleanup, PARTICLEANUP)

/* added for fortran */
#ifdef LAMMPI
extern void** lam_F_handles; /* found in LAM */
void PARTISETUP_F77(int* grp) {
  (void)PARTI_MPI_setup((MPI_Comm*)&lam_F_handles[*grp]);
}
#else
void PARTISETUP_F77(int* grp) {
  (void)PARTI_MPI_setup((MPI_Comm*)grp);
}
#endif

/* added for fortran */
void PARTICLEANUP_F77() {
  /* destroy all the fortran objects */
  destroy_sch_table();
  destroy_da_table();
  destroy_decomp_table();
  destroy_vproc_table();
  /* destroy all the object caches */
  PARTI_MPI_cleanup();
}

#endif /* USE_MPI */

#endif /* FORTRAN */
