/** 
 * The public interface of MultiBlockParti
 * @file multiblockparti.h
 * @author Christian Hansen (chansen@cs.umd.edu)
 *         University of Maryland, College Park
 */

#ifndef MULTIBLOCKPARTI_H
#define MULTIBLOCKPARTI_H

#include "port.h"

#define MAX_DIM   6
#define MAX_NODES 1024

typedef struct decomp_block {
  int task;               /* owner */
  int rank;               /* MC compatability */
  int* size;              /* local */
  int* globalLowerBound;  
  int* globalUpperBound;
  int* coord;             /* in decomp space */
} decomp_block_t;

typedef struct decomp_tree {
  int dimension;                /* dimension of this partition */
  int partitions;               /* number of partitions */
  struct decomp_tree* children; /* partitions */
  int* keys;                    /* keys for identifying partitions */
  decomp_block_t* leaf;         /* block information if leaf */
} decomp_tree_t;

/** A block irregular decomposition */
typedef struct parti_decomp {
  int rank;             /* number of dimensions */
  int* size;            /* size of each dimension */
  decomp_tree_t root;   /* decomposition structure */
  int referenceCount;
  void* temp;           /* temporary pointer an old style decomp */
} parti_decomp_t;

/** A distributed array descriptor */
typedef struct parti_darray {
  int* lowerGhostCells;    /* internal ghost cells */
  int* upperGhostCells;    /* internal ghost cells */
  parti_decomp_t* decomp; 
  decomp_block_t* block;   /* block owned by this task */
  int referenceCount;
} parti_darray_t;

/** An array region */
typedef struct parti_region {
  int rank;
  int* lower;
  int* upper;
  int* stride;
} parti_region_t;

typedef struct sched_data {
  int proc;             /* Processor Number */
  int numDims;          /* No. of dimensions in the arrays involved */
  int startPosn;        /* Starting position of the very first element */
  int numelem[MAX_DIM]; 
  int str[MAX_DIM];     /* Stride in each dimension includes the multiplier factor */  
} SchedData; 

/** A communication schedule */
typedef struct parti_sched {
  int rMsgSz[MAX_NODES];
  int sMsgSz[MAX_NODES];
  SchedData* rData[MAX_NODES];
  SchedData* sData[MAX_NODES];
  int type;            /* 1 for exch_sched, 2 for subarray_sched */
  int hash;            /* bucket in the hash table */
  int referenceCount;
  int ID_Number;
} parti_sched_t;

/**
 * Create an irregular block decomposition.
 * @param rank  the rank
 * @param size  the size of each dimension
 * @return the irregular block decomposition
 */
parti_decomp_t* parti_create_idecomp(int rank, int* size);

/**
 * Create all the partitions for all the dimensions.
 * @param decomp  the decomposition
 * @param dim     the dimensions to split
 * @param cnt     the number of partitions to create for each dimension
 * @param idx     the limiting upper index of each partition for each dimension
 * @return the number of blocks created
 */
int parti_section(parti_decomp_t* decomp, int* dim, int* cnt, int** idx);

/**
 * Create a partition in the decomposition.
 * @param decomp  the decomposition
 * @param dim     the dimension to split
 * @param blk     the block to split
 * @param cnt     the number of partitions to create
 * @param idx     the limiting upper index of each partition
 * @return the number of partitions created
 */
int parti_partition(parti_decomp_t* decomp, int dim, int* blk, int cnt, int* idx);

/**
 * Create a distributed array descriptor.
 * @param decomp  the decomposition to use
 * @param igcl    the number of internal ghost cells on the lower end
 * @param igcu    the number of internal ghost cells on the upper end
 * @return the distributed array descriptor
 */
parti_darray_t* parti_ialign(parti_decomp_t* decomp, int* igcl, int* igcu);

#define PARTI_OUT_OF_RANGE -99

/**
 * Provides the lowest/greatest valid local index for a global index.
 * @param darray  a distributed array descriptor
 * @param dim     a dimension
 * @param idx     the global index
 * @param stride  the stride
 * @return the local index
 */
int parti_local_loop_bound(parti_darray_t* darray, int dim, int idx, int stride);

/**
 * Provide the size of the local portion of the array.
 * @param darray  a distributed array descriptor
 * @param size    optional receptacle for the size of each dimension
 * @return the total size of the local array
 */
int parti_local_size(parti_darray_t* darray, int* size);

/**
 * Provide the lower global bound for the local block.
 * @param darray  a distributed array descriptor
 * @param dim     a dimension
 * @return the global bound
 */
int parti_global_lower_bound(parti_darray_t* darray, int dim);

/**
 * Provide the upper global bound for the local block.
 * @param darray  a distributed array descriptor
 * @param dim     a dimension
 * @return the global bound
 */
int parti_global_upper_bound(parti_darray_t* darray, int dim);

/**
 * Convert a global index to a local index.
 * @param darray  a distributed array descriptor
 * @param idx     a global index
 * @param dim     a dimension
 * @param ghost   a flag indicating whether or not to consider internal ghost cells
 * @return the local index
 */
int parti_global_to_local(parti_darray_t* darray, int idx, int dim, int ghost);

/**
 * Converts local index to a global index.
 * @param darray  a distributed array descriptor
 * @param idx     a local index
 * @param dim     a dimension
 * @param ghost   a flag indicating whether or not to consider internal ghost cells
 */
int parti_local_to_global(parti_darray_t* darray, int idx, int dim, int ghost);

/**
 * Create a schedule for filling only the ghost cells specified.
 * @param darray  a distributed array descriptor
 * @param fill    a vector explicitly indicating the ghost cells to fill
 * @return a communication schedule
 */
parti_sched_t* parti_ghost_sched_vector(parti_darray_t* darray, int* fill);

/**
 * Create a schedule for filling all the ghost cells up to and including
 * the fill vector specified.
 * @param darray  a distributed array descriptor
 * @param fill    a vector indicating a range of ghost cells to fill
 * @return a communication schedule
 */
parti_sched_t* parti_ghost_sched_span(parti_darray_t* darray, int* fill);

/**
 * Create a schedule for filling all the ghost cells of an array.
 * @param darray  a distributed array descriptor
 * @return a communication schedule
 */
parti_sched_t* parti_ghost_sched_all(parti_darray_t* darray);

/**
 * Create a schedule for filling all the ghost cells of a single dimension.
 * @param darray  a distributed array descriptor
 * @param dim     the ghost cell dimension
 * @param fill    the number of ghost cells to fill
 * @return a communication schedule
 */
parti_sched_t* parti_ghost_sched_abs(parti_darray_t* darray, int dim, int fill);

/**
 * Creates an array region structure for defining communication schedules.
 * @param rank  the dimension of the region
 * @return a region structure
 */
parti_region_t* parti_alloc_region(int rank);

/**
 * Computes a schedule transferring a region in one array to a region in another array,
 * optionally re-aligning the sender's dimensions to differnt receiver's dimensions using 
 * a map.
 * @param srceArray   a distributed array descriptor
 * @param srceRegion  a region definition
 * @param destArray   a distributed array descriptor
 * @param destRegion  a region definition
 * @param map         the map between dimensions
 * @return a communication schedule
 */
parti_sched_t* parti_region_sched(parti_darray_t* srceArray, parti_region_t* srceRegion, 
				  parti_darray_t* destArray, parti_region_t* destRegion, int* map);

/**
 * Sends and receives data based on the given communication schedule.
 * @param srce   the local memory of the source array
 * @param sched  the communication schedule
 * @param dest   the local memory of the destination array
 */
void parti_data_move_char(char* srce, parti_sched_t* sched, char* dest);

/**
 * Sends and receives data based on the given communication schedule.
 * @param srce   the local memory of the source array
 * @param sched  the communication schedule
 * @param dest   the local memory of the destination array
 */
void parti_data_move_int(int* srce, parti_sched_t* sched, int* dest);

/**
 * Sends and receives data based on the given communication schedule.
 * @param srce   the local memory of the source array
 * @param sched  the communication schedule
 * @param dest   the local memory of the destination array
 */
void parti_data_move_float(float* srce, parti_sched_t* sched, float* dest);

/**
 * Sends and receives data based on the given communication schedule.
 * @param srce   the local memory of the source array
 * @param sched  the communication schedule
 * @param dest   the local memory of the destination array
 */
void parti_data_move_double(double* srce, parti_sched_t* sched, double* dest);

/**
 * Frees a communication schedule.
 * @param sched  the communication schedule
 */
void parti_free_sched(parti_sched_t* sched);

/**
 * Frees an array region.
 * @param region  the array region
 */
void parti_free_region(parti_region_t* region);

/** 
 * Frees an irregular block decomposition.
 * @param decomp  the irregular block decomposition
 */
void parti_free_decomp(parti_decomp_t* decomp);

/** 
 * Frees a distributed array descriptor.
 * @param darray  the distributed array descriptor
 */
void parti_free_darray(parti_darray_t* darray);

/**
 * Used before all other MultiBlockParti calls, this function
 * takes either a MPI communicator or a PVM group name for internal
 * communication use.
 */
#define parti_setup PARTI_setup

/**
 * Used after all other MultiBlockParti calls, this function
 * clears the cached communication schedules and other internal
 * memory structures.
 */
#define parti_cleanup PARTI_cleanup

#endif /* MULTIBLOCKPARTI_H */
