/* 
 * Auth: Christian Hansen (chansen@cs.umd.edu)
 *       University of Maryland, College Park
 * Desc: Routines for creating communication schedules to
 *       send and receive ghost cell data.
 */

#include "schedules.h"
#include "birreg.h"
#include "decomp.h"
#include "exchsched.h"
#include "hash.h"
#include "list.h"
#include "utils.h"
#include "addition.h"
#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define min(A,B) ((A)<(B)?(A):(B))
#define max(A,B) ((A)>(B)?(A):(B))

/* allocates a region */
parti_region_t* parti_alloc_region(int rank) {
  parti_region_t* region;

  region = (parti_region_t*)malloc(sizeof(parti_region_t));
  region->rank = rank;
  region->lower = (int*)calloc(rank, sizeof(int));
  region->upper = (int*)calloc(rank, sizeof(int));
  region->stride = (int*)calloc(rank, sizeof(int));

  memset(region->lower, 0, rank*sizeof(int));
  memset(region->upper, 0, rank*sizeof(int));
  memset(region->stride, 0, rank*sizeof(int));

  return region;
}

/* copies a region */
parti_region_t* copy_region(parti_region_t* source) {
  parti_region_t* region = 0;
  int i;

  if (source) {
    region = parti_alloc_region(source->rank);
    for (i = 0; i < source->rank; ++i) {
      region->lower[i] = source->lower[i];
      region->upper[i] = source->upper[i];
      region->stride[i] = source->stride[i];
    }
  }

  return region;
}

/* deallocates a region */
void parti_free_region(parti_region_t* region) {
  if (region) {
    if (region->lower) free(region->lower);
    if (region->upper) free(region->upper);
    if (region->stride) free(region->stride);
    free(region);
  }
}

/* prints a region */
void print_region(parti_region_t* region) {
  int i;

  printf("<region>\n");
  printf("  rank=%d\n", region->rank);
  for (i = 0; i < region->rank; ++i) {
    printf("    lo[%d]=%d\n", i, region->lower[i]);
    printf("    up[%d]=%d\n", i, region->upper[i]);
    printf("    st[%d]=%d\n", i, region->stride[i]);
  }
  printf("</region>\n");
}

/* prints a schedule element */
void print_sched_data(char* title, SchedData* data) {
  int i;

  printf("%s\n", title);
  printf("  proc=%d\n", data->proc);
  printf("  rank=%d\n", data->numDims);
  printf("  offs=%d\n", data->startPosn);
  for (i = 0; i < data->numDims; ++i) {
    printf("  elem[%d]=%d\n", i, data->numelem[i]);
    printf("  strd[%d]=%d\n", i, data->str[i]);
  }
}

/* void print_sched(parti_sched_t* sched) { */
/*   int i; */

/*   if (sched) { */
/*     printf("-SCHED-\n"); */
/*     for (i = 0; i < MAX_NODES; ++i) { */
/*       if (sched->rMsgSz[i] > 0) { */
/* 	print_sched_data("receiving...", sched->rData[i]); */
/*       } */
/*       if (sched->sMsgSz[i] > 0) { */
/* 	print_sched_data("sending...", sched->sData[i]); */
/*       } */
/*     } */
/*     printf("type=%d\n", sched->type); */
/*     printf("hash=%d\n", sched->hash); */
/*     printf("ID=%d\n", sched->ID_Number); */
/*   } */
/* } */

/* compute the memory distance between consecutive elements in a block */
int compute_stride(parti_darray_t* darray, int dim) {
  int i, stride = 1;
  
  if (FFlag) {
    for (i = 0; i < dim; ++i) 
      stride *= (darray->block->size[i] + darray->lowerGhostCells[i] + darray->upperGhostCells[i]);
  } else {
    for (i = darray->decomp->rank - 1; i > dim; --i)
      stride *= (darray->block->size[i] + darray->lowerGhostCells[i] + darray->upperGhostCells[i]);
  }
  
  return stride;
}

/* compute the memory offset for the first element of the region */
int compute_offset(parti_darray_t* darray, parti_region_t* region) {
  int i, offset = 0;
  
  if (FFlag) {
    for (i = 0; i < darray->decomp->rank; ++i) {
      offset += region->lower[i]*region->stride[i];
    }
  } else {
    for (i = darray->decomp->rank - 1; i >= 0; --i) {
      offset += region->lower[i]*region->stride[i];
    }
  }

  return offset;
}

int get_intersection(int rank, parti_region_t* region, decomp_block_t* block, parti_region_t* result) {
  int i;

  for (i = 0; i < rank; ++i) {
    if (region->lower[i] <= block->globalUpperBound[i] && region->upper[i] >= block->globalLowerBound[i]) {
      result->lower[i] = region->lower[i];
      while (result->lower[i] < block->globalLowerBound[i])
	result->lower[i] += region->stride[i];
      if (result->lower[i] > block->globalUpperBound[i])
	return 0;
      result->upper[i] = region->upper[i];
      while (result->upper[i] > block->globalUpperBound[i])
	result->upper[i] -= region->stride[i];
      if (result->upper[i] < block->globalLowerBound[i])
	return 0;
    } else {
      return 0;
    }
    result->stride[i] = region->stride[i];
  }

  for (i = rank; i < region->rank; ++i) {
    result->lower[i] = 0;
    result->upper[i] = 0;
    result->stride[i] = 1;
  }
  
  return 1;
}

/* projects the subregion (s1) of one region (r1) into another region (r2) to create subregion (s2) */
void get_projection(int rank, parti_region_t* r1, parti_region_t* r2, 
		    parti_region_t* s1, parti_region_t* s2, int* map) {
  int i;
  
  for (i = 0; i < rank; ++i) {
    s2->lower[map[i]] = ((s1->lower[i] - r1->lower[i])/r1->stride[i])*r2->stride[map[i]] + r2->lower[map[i]];
    s2->upper[map[i]] = ((s1->upper[i] - r1->lower[i])/r1->stride[i])*r2->stride[map[i]] + r2->lower[map[i]];
    s2->stride[map[i]] = r2->stride[map[i]];
  }
}

parti_sched_t* parti_ghost_sched_vector(parti_darray_t* darray, int* fill) {
  parti_sched_t* sched = 0;
  parti_region_t* globalRegion;
  int i;

  decomp_block_t** blocks = 0;
  SchedData* recvData = 0;
  SchedData* sendData = 0;
  parti_region_t* localRegion;
  parti_region_t* otherRegion;
  int j, count;
  int totalSize;

  if (darray == 0 || darray->decomp == 0 || darray->block == 0)
    return sched;
  
  /* check if requested fill exceeds allocated ghost cells */
  for (i = 0; i < darray->decomp->rank; ++i) {
    if (fill[i] < -darray->lowerGhostCells[i] || fill[i] > darray->upperGhostCells[i]) {
      fprintf(stderr, "parti_ghost_sched_vector: fill exceeded allocated ghost cells\n");
      return sched;
    }
  }
  
  /* check to see if there is already a schedule for this request */
  sched = lookup_exch((DARRAY*)darray, darray->decomp->rank, fill); /* cast for the compiler */
  if (sched) {
#if PARTI_DEBUG > 2
    printf("Node %d: reusing saved exchange schedule\n", PARTI_myproc());
#endif
    return sched;
  }

  /* create a new inter-array schedule */
  sched = (parti_sched_t*)pmalloc("parti_sched_t", sizeof(parti_sched_t));
  init_Sched(sched);
  sched->type = 1;

  /* determine the global region for the requested ghost cells */
  globalRegion = parti_alloc_region(darray->decomp->rank);
  for (i = 0; i < darray->decomp->rank; ++i) {
    if (fill[i] < 0) {
      globalRegion->lower[i] = darray->block->globalLowerBound[i] + fill[i];
      globalRegion->upper[i] = darray->block->globalLowerBound[i] - 1;
    } else if (fill[i] > 0) {
      globalRegion->lower[i] = darray->block->globalUpperBound[i] + 1;
      globalRegion->upper[i] = darray->block->globalUpperBound[i] + fill[i];
    } else {
      globalRegion->lower[i] = darray->block->globalLowerBound[i];
      globalRegion->upper[i] = darray->block->globalUpperBound[i];
    }
  }
#if PARTI_DEBUG > 1
  print_region(globalRegion);
#endif
  
  blocks = 0;
  count = 0;
  
  get_region_blocks(globalRegion, &(darray->decomp->root), &blocks, &count);
  for (i = 0; i < count; ++i) {
    if (blocks[i] == darray->block)
      continue;

    recvData = (SchedData*)malloc(sizeof(SchedData));
    recvData->proc = blocks[i]->task;
    recvData->numDims = darray->decomp->rank;
    totalSize = 1;

    /* define the local region for this remote block */
    localRegion = parti_alloc_region(darray->decomp->rank);
    for (j = 0; j < darray->decomp->rank; ++j) {
      localRegion->lower[j] = max(blocks[i]->globalLowerBound[j], globalRegion->lower[j]) 
	- darray->block->globalLowerBound[j] + darray->lowerGhostCells[j];
      localRegion->upper[j] = min(blocks[i]->globalUpperBound[j], globalRegion->upper[j])
	- darray->block->globalLowerBound[j] + darray->lowerGhostCells[j];
      recvData->numelem[j] = localRegion->upper[j] - localRegion->lower[j] + 1;
      totalSize *= recvData->numelem[j];
      recvData->str[j] = localRegion->stride[j] = compute_stride(darray, j);
    }
#if PARTI_DEBUG > 1
    print_region(localRegion);
#endif

    recvData->startPosn = compute_offset(darray, localRegion);
    parti_free_region(localRegion);

    sched->rData[blocks[i]->task] = recvData;
    sched->rMsgSz[blocks[i]->task] = totalSize;
#if PARTI_DEBUG > 1
    print_sched_data("RECV SCHEDULE", recvData);
    printf("totalSize=%d\n", totalSize);
#endif
  }  

  free(blocks);

  /* determine the global region which neighbors the requested ghostcells */
  for (i = 0; i < darray->decomp->rank; ++i) {
    if (fill[i] < 0) {
      globalRegion->lower[i] = darray->block->globalLowerBound[i] + 1;
      globalRegion->upper[i] = darray->block->globalUpperBound[i] - fill[i];
    } else if (fill[i] > 0) {
      globalRegion->lower[i] = darray->block->globalLowerBound[i] - fill[i];
      globalRegion->upper[i] = darray->block->globalUpperBound[i] - 1;
    } else {
      globalRegion->lower[i] = darray->block->globalLowerBound[i];
      globalRegion->upper[i] = darray->block->globalUpperBound[i];
   }
  }
#if PARTI_DEBUG > 1
  print_region(globalRegion);
#endif
  
  blocks = 0;
  count = 0;

  get_region_blocks(globalRegion, &(darray->decomp->root), &blocks, &count);
  for (i = 0; i < count; ++i) {
    if (blocks[i] == darray->block)
      continue;

    /* find the ghost cell region for the other blocks */
    otherRegion = parti_alloc_region(darray->decomp->rank);
    for (j = 0; j < darray->decomp->rank; ++j) {
      if (fill[j] < 0) {
	otherRegion->lower[j] = blocks[i]->globalLowerBound[j] + fill[j];
	otherRegion->upper[j] = blocks[i]->globalLowerBound[j] - 1;
      } else if (fill[j] > 0) {
	otherRegion->lower[j] = blocks[i]->globalUpperBound[j] + 1;
	otherRegion->upper[j] = blocks[i]->globalUpperBound[j] + fill[j];
      } else {
	otherRegion->lower[j] = blocks[i]->globalLowerBound[j];
	otherRegion->upper[j] = blocks[i]->globalUpperBound[j];
      }
      otherRegion->stride[j] = 1;
    }
  
    localRegion = parti_alloc_region(darray->decomp->rank);    
    if (!get_intersection(darray->decomp->rank, otherRegion, darray->block, localRegion)) {
      parti_free_region(localRegion);
      parti_free_region(otherRegion);
      continue;
    }
    
    sendData = (SchedData*)malloc(sizeof(SchedData));
    sendData->proc = blocks[i]->task;
    sendData->numDims = darray->decomp->rank;
    totalSize = 1;  

    for (j = 0; j < darray->decomp->rank; ++j) {
      localRegion->lower[j] = localRegion->lower[j] - darray->block->globalLowerBound[j] + darray->lowerGhostCells[j];
      localRegion->upper[j] = localRegion->upper[j] - darray->block->globalLowerBound[j] + darray->lowerGhostCells[j];
      sendData->numelem[j] = localRegion->upper[j] - localRegion->lower[j] + 1;
      totalSize *= sendData->numelem[j];
      sendData->str[j] = localRegion->stride[j] = compute_stride(darray, j);
    }
    sendData->startPosn = compute_offset(darray, localRegion);
    parti_free_region(localRegion);
    parti_free_region(otherRegion);

    sched->sData[blocks[i]->task] = sendData;
    sched->sMsgSz[blocks[i]->task] = totalSize;
#if PARTI_DEBUG > 1
    print_sched_data("SEND SCHEDULE", sendData);
    printf("totalSize=%d\n", totalSize);
#endif
  }

  free(blocks);  
  parti_free_region(globalRegion);

  /* save the generated schedule in the hash table */
#if PARTI_DEBUG > 2
  printf("Node %d: saving exchange schedule\n", PARTI_myproc());
#endif
  insert_exch_table(sched, darray, darray->decomp->rank, fill); 
  
  return sched;
}

parti_sched_t* parti_ghost_sched_span(parti_darray_t* darray, int* fill) {
  parti_sched_t* sched = 0;
  parti_sched_t* partialSched;
  int i, cnt = 0, valid = 0;
  list combined;
  int* partialFill;
  int* item;

  if (darray == 0 || darray->decomp == 0 || darray->block == 0)
    return sched;
  
  /* check if requested fill exceeds allocated ghost cells */
  for (i = 0; i < darray->decomp->rank; ++i) {
    if (fill[i] < -darray->lowerGhostCells[i] || fill[i] > darray->upperGhostCells[i]) {
      fprintf(stderr, "parti_ghost_sched_span: fill exceeded allocated ghost cells\n");
      return sched;
    }
  }

  sched = (parti_sched_t*)pmalloc("parti_sched_t", sizeof(parti_sched_t));
  init_Sched(sched);
  
  combined = create_list();
 
  /* count the non-zeros */
  for (i = 0; i < darray->decomp->rank; ++i) {
    if (fill[i] != 0) {
      ++cnt;
    }
  }

  /* generate all combinations of fills */
  for (i = 0; i < cnt; ++i) {
    partialFill = (int*)calloc(darray->decomp->rank, sizeof(int));
    comb(darray->decomp->rank, i+1, partialFill, 0, combined);
  }

  /* create schedules for each valid fill and merge into schedule */
  FOREACH (item, combined) {
    valid = 1;
    for (i = 0; i < darray->decomp->rank; ++i) {
      if (item[i] == 1 && fill[i] == 0)
	valid = 0;
      item[i] *= fill[i];
    }
    if (valid) {
      partialSched = parti_ghost_sched_vector(darray, item);
      merge_sched(sched, partialSched);
      delete_SCHED(partialSched);
    }    
    free(item);
  }  

  destroy_list(combined);

#if PARTI_DEBUG > 1
  for (i = 0; i < PARTI_numprocs(); ++i) {
    if (sched->rMsgSz[i] > 0) {
      print_sched_data("MERGED RECV SCHEDULE", sched->rData[i]);
      printf("rMsgSz=%d\n", sched->rMsgSz[i]);
    }
    if (sched->sMsgSz[i] > 0) {
      print_sched_data("MERGED SEND SCHEDULE", sched->sData[i]);
      printf("sMsgSz=%d\n", sched->sMsgSz[i]);
    }
  }
#endif
  
  return sched;
}

parti_sched_t* parti_ghost_sched_all(parti_darray_t* darray) {
  parti_sched_t* sched = 0;
  parti_sched_t* partialSched;
  list permutations;
  int* lowerFill, *upperFill;
  int* partialFill;
  int* item;
  int i;

  if (darray == 0 || darray->decomp == 0 || darray->block == 0)
    return sched;
  
  lowerFill = (int*)calloc(darray->decomp->rank, sizeof(int));
  upperFill = (int*)calloc(darray->decomp->rank, sizeof(int));
  for (i = 0; i < darray->decomp->rank; ++i) {
    lowerFill[i] = -darray->lowerGhostCells[i];
    upperFill[i] = darray->upperGhostCells[i];
  }

  sched = (parti_sched_t*)pmalloc("parti_sched_t", sizeof(parti_sched_t));
  init_Sched(sched);

  permutations = create_list();
  
  /* generate all combinations of fill vectors */
  partialFill = (int*)calloc(darray->decomp->rank, sizeof(int));
  perm(darray->decomp->rank, lowerFill,  upperFill, partialFill, 0, permutations);
  free(partialFill);
  partialFill = remove_list(0, permutations);
  
  FOREACH (item, permutations) {
    partialSched = parti_ghost_sched_vector(darray, item);
    merge_sched(sched, partialSched);
    delete_SCHED(partialSched);
    free(item);
  }
  destroy_list(permutations);
  
  free(lowerFill);
  free(upperFill);

#if PARTI_DEBUG > 1
  for (i = 0; i < PARTI_numprocs(); ++i) {
    if (sched->rMsgSz[i] > 0) {
      print_sched_data("MERGED RECV SCHEDULE", sched->rData[i]);
      printf("rMsgSz=%d\n", sched->rMsgSz[i]);
    }
    if (sched->sMsgSz[i] > 0) {
      print_sched_data("MERGED SEND SCHEDULE", sched->sData[i]);
      printf("sMsgSz=%d\n", sched->sMsgSz[i]);
    }
  }
#endif

  return sched;
}

parti_sched_t* parti_ghost_sched_abs(parti_darray_t* darray, int dim, int fill) {
  parti_sched_t* sched = 0;
  int i;
  int* totalFill;
  
  if (darray == 0 || darray->decomp == 0 || darray->block == 0)
    return sched;
  
  /* check if requested fill exceeds allocated ghost cells */
  if (fill < -darray->lowerGhostCells[dim] || fill > darray->upperGhostCells[dim]) {
    fprintf(stderr, "parti_ghost_sched_abs: fill exceeded allocated ghost cells\n");
    return sched;
  }

  totalFill = (int*)calloc(darray->decomp->rank, sizeof(int));
  for (i = 0; i < darray->decomp->rank; ++i) {
    if (i == dim)
      totalFill[i] = fill;
    else
      totalFill[i] = 0;
  }
  sched = parti_ghost_sched_vector(darray, totalFill);
  free(totalFill);

  return sched;
}

parti_region_t* resize_region(int rank, parti_region_t* source) {
  parti_region_t* region;
  int i;

  region = parti_alloc_region(rank);
  for (i = 0; i < source->rank; ++i) {
    region->lower[i] = source->lower[i];
    region->upper[i] = source->upper[i];
    region->stride[i] = source->stride[i];
  }
  for (i = source->rank; i < rank; ++i) {
    region->lower[i] = 0;
    region->upper[i] = 0;
    region->stride[i] = 1;
  }
  
  return region;
}

int* remap(int sr, int* old, int dr) {
  int i, j, k, u = 0, *map;

  if (sr > dr) {
    map = (int*)calloc(sr, sizeof(int));
    k = dr;
    for (i = 0; i < sr; ++i)
      if (old[i] != -1) 
	map[i] = old[i];
      else
	map[i] = k++;
  } else if (dr > sr) {
    map = (int*)calloc(dr, sizeof(int));
    for (i = 0; i < sr; ++i)
      map[i] = old[i];
    for (j = 0; j < dr; ++j) {
      for (k = 0; k < sr; ++k)
	if (map[k] == j)
	  u = 1;
      if (u == 0)
	map[i++] = j;
      else
	u = 0;
    }
  } else {
    map = (int*)calloc(sr, sizeof(int));
    for (i = 0; i < sr; ++i)
      map[i] = old[i];
  }

  return map;
}

parti_sched_t* parti_region_sched(parti_darray_t* srceArray, parti_region_t* srceRegion,
				  parti_darray_t* destArray, parti_region_t* destRegion,
				  int* dimensionMap) {
  parti_sched_t* sched = 0;
  parti_region_t* srceRegionCopy, *destRegionCopy;
  parti_region_t* srceRegionLocal, *destRegionLocal;
  int i, j, k, rank, *map, *inv;
  int srceSize, destSize;
  int sending = 1, receiving = 1;
  int* tmp1, *tmp2;

  decomp_block_t** blocks = 0;
  SchedData* sendData = 0;
  SchedData* recvData = 0;
  parti_region_t* srceSubregionLocal, *destSubregionLocal;
  int count;
  int totalSize;

  /* check that we have valid parameters */
  if (srceArray == 0 || destArray == 0 || srceRegion == 0 || destRegion == 0 || dimensionMap == 0) {
    return sched;
  }

  /* check that we have valid array descriptors */
  if (srceArray->decomp == 0 || destArray->decomp == 0) {
    return sched;
  }

  /* check that we have useful parameters for this call */
  if (srceArray->decomp->rank != srceRegion->rank || destArray->decomp->rank != destRegion->rank) {
    return sched;
  }

  /* see if there is an existing schedule */
  sched = lookup_sub(srceArray, destArray, min(srceArray->decomp->rank, destArray->decomp->rank),
		     dimensionMap, srceRegion->lower, srceRegion->upper, srceRegion->stride,
		     dimensionMap, destRegion->lower, destRegion->upper, destRegion->stride);
  if (sched) {
#if PARTI_DEBUG > 2
    printf("Node %d: reusing saved exchange schedule\n", PARTI_myproc());
#endif
    return sched;
  }

  /* allow us to act like the arrays are of the same rank */
  if (srceArray->decomp->rank < destArray->decomp->rank) {
    rank = destArray->decomp->rank;
    srceRegionCopy = resize_region(rank, srceRegion);
    destRegionCopy = copy_region(destRegion);
    map = remap(srceArray->decomp->rank, dimensionMap, destArray->decomp->rank);
  } else if (destArray->decomp->rank < srceArray->decomp->rank) {
    rank = srceArray->decomp->rank;
    srceRegionCopy = copy_region(srceRegion);
    destRegionCopy = resize_region(rank, destRegion);
    map = remap(srceArray->decomp->rank, dimensionMap, destArray->decomp->rank);
  } else {
    rank = srceArray->decomp->rank;
    srceRegionCopy = copy_region(srceRegion);
    destRegionCopy = copy_region(destRegion);
    map = remap(srceArray->decomp->rank, dimensionMap, destArray->decomp->rank);
  }

#if PARTI_DEBUG > 1
  printf("SRCE REGION COPY\n");
  print_region(srceRegionCopy);
  printf("DEST REGION COPY\n");
  print_region(destRegionCopy);
  printf("MAPPING\n");
  for (i = 0; i < rank; ++i)
    printf("%d ", map[i]);
  printf("\n");
#endif

  /* check that source region fits into the destination region for both size and shape */
  for (i = 0; i < rank; ++i) {
    srceSize = (srceRegionCopy->upper[i] - srceRegionCopy->lower[i])/srceRegionCopy->stride[i];
    destSize = (destRegionCopy->upper[map[i]] - destRegionCopy->lower[map[i]])/destRegionCopy->stride[map[i]];
    if (srceSize != destSize) {
      fprintf(stderr, "parti_region_sched: source and destination regions do not match\n");
      parti_free_region(srceRegionCopy);
      parti_free_region(destRegionCopy);
      free(map);
      return sched;
    }
  }
  
  /* create an inverted mapping for later calculations */
  inv = (int*)calloc(rank, sizeof(int));
  for (i = 0; i < rank; ++i)
    inv[map[i]] = i;

  /* create a new intra-array schedule */
  sched = (parti_sched_t*)pmalloc("parti_sched_t", sizeof(parti_sched_t));
  init_Sched(sched);
  sched->type = 2;
  
  /* find the intersection of the source region with the local array */
  srceRegionLocal = parti_alloc_region(rank);
  if (!srceArray->block)
    sending = 0;
  else
    sending = get_intersection(srceArray->decomp->rank, srceRegionCopy, srceArray->block, srceRegionLocal);
#if PARTI_DEBUG > 1
  printf("SRCE REGION LOCAL\n");
  print_region(srceRegionLocal);
#endif
  
  /* project the local portion of the source region onto the destination array */
  destRegionLocal = parti_alloc_region(rank);
  get_projection(rank, srceRegionCopy, destRegionCopy, srceRegionLocal, destRegionLocal, map);
#if PARTI_DEBUG > 1
  printf("DEST REGION LOCAL\n");
  print_region(destRegionLocal);
#endif
  
  if (sending) {
    blocks = 0;
    count = 0;
    
    /* determine who owns this portion of destination region and create a send schedule */
    get_region_blocks(destRegionLocal, &(destArray->decomp->root), &blocks, &count);
    for (i = 0; i < count; ++i) {
      totalSize = 1;
      sendData = (SchedData*)malloc(sizeof(SchedData));
      sendData->proc = blocks[i]->task;
      sendData->numDims = rank;
      destSubregionLocal = parti_alloc_region(rank);

      /* find the intersection of the destination region and the entire block owned by this task */
      sending = get_intersection(destArray->decomp->rank, destRegionLocal, blocks[i], destSubregionLocal);      
#if PARTI_DEBUG > 1
      printf("DEST SUBREGION LOCAL\n");
      print_region(destSubregionLocal);
#endif 
      if (sending) {
	srceSubregionLocal = parti_alloc_region(rank);
	get_projection(rank, destRegionLocal, srceRegionLocal, destSubregionLocal, srceSubregionLocal, inv);
#if PARTI_DEBUG > 1
	printf("SRCE SUBREGION LOCAL\n");
	print_region(srceSubregionLocal);
#endif
	for (j = 0; j < srceArray->decomp->rank; ++j) {
	  /* convert this region into local coordinates */
	  srceSubregionLocal->lower[j] = srceSubregionLocal->lower[j] - srceArray->block->globalLowerBound[j] + srceArray->lowerGhostCells[j];
	  srceSubregionLocal->upper[j] = srceSubregionLocal->upper[j] - srceArray->block->globalLowerBound[j] + srceArray->lowerGhostCells[j];
	  sendData->numelem[j] = (srceSubregionLocal->upper[j] - srceSubregionLocal->lower[j])/srceSubregionLocal->stride[j] + 1;
	  totalSize *= sendData->numelem[j];
	  sendData->str[j] = srceSubregionLocal->stride[j] = compute_stride(srceArray, j);
	  sendData->str[j] *= srceRegionLocal->stride[j];
	}
	for (j = srceArray->decomp->rank; j < rank; ++j) {
	  sendData->numelem[j] = sendData->str[j] = 1;
	}
	sendData->startPosn = compute_offset(srceArray, srceSubregionLocal);
	parti_free_region(srceSubregionLocal);

	sched->sData[blocks[i]->task] = sendData;
	sched->sMsgSz[blocks[i]->task] = totalSize;
#if PARTI_DEBUG > 1
	print_sched_data("SEND SCHEDULE", sendData);
	printf("totalSize=%d\n", totalSize);
#endif
      }
      parti_free_region(destSubregionLocal);
    }    
    free(blocks);
  }

  /* find the intersection of the destination region with the local array */
  if (!destArray->block)
    receiving = 0;
  else 
    receiving = get_intersection(destArray->decomp->rank, destRegionCopy, destArray->block, destRegionLocal);
#if PARTI_DEBUG > 1
  printf("DEST REGION LOCAL\n");
  print_region(destRegionLocal);
#endif
  
  /* project the local portion of the destination region back into the source array */
  get_projection(rank, destRegionCopy, srceRegionCopy, destRegionLocal, srceRegionLocal, inv);
#if PARTI_DEBUG > 1
  printf("SRCE REGION LOCAL\n");
  print_region(srceRegionLocal);
#endif

  if (receiving) {
    blocks = 0;
    count = 0;
    
    /* determine who owns this portion of the source region and create a receive schedule */
    get_region_blocks(srceRegionLocal, &(srceArray->decomp->root), &blocks, &count);
    for (i = 0; i < count; ++i) {
      totalSize = 1;
      recvData = (SchedData*)malloc(sizeof(SchedData));
      recvData->proc = blocks[i]->task;
      recvData->numDims = rank;
      srceSubregionLocal = parti_alloc_region(rank);

      /* find the intersection of the source region and the entire block owned by this task */
      receiving = get_intersection(srceArray->decomp->rank, srceRegionLocal, blocks[i], srceSubregionLocal);
#if PARTI_DEBUG > 2
      printf("SRCESUBREGION LOCAL\n");
      print_region(srceSubregionLocal);
#endif
      if (receiving) {
	destSubregionLocal = parti_alloc_region(rank);
	get_projection(rank, srceRegionLocal, destRegionLocal, srceSubregionLocal, destSubregionLocal, map);
#if PARTI_DEBUG > 2
	printf("DESTSUBREGION LOCAL\n");
	print_region(destSubregionLocal);
#endif
	for (j = 0; j < destArray->decomp->rank; ++j) {
	  /* convert this region into local coordinates */
	  destSubregionLocal->lower[j] = destSubregionLocal->lower[j] - destArray->block->globalLowerBound[j] + destArray->lowerGhostCells[j];
	  destSubregionLocal->upper[j] = destSubregionLocal->upper[j] - destArray->block->globalLowerBound[j] + destArray->lowerGhostCells[j];
	  recvData->numelem[inv[j]] = (destSubregionLocal->upper[j] - destSubregionLocal->lower[j])/destRegionLocal->stride[j] + 1;
	  totalSize *= recvData->numelem[inv[j]];
	  recvData->str[inv[j]] = destSubregionLocal->stride[j] = compute_stride(destArray, j);
	  recvData->str[inv[j]] *= destRegionLocal->stride[j];
	}
	for (j = destArray->decomp->rank; j < rank; ++j) {
	  recvData->numelem[inv[j]] = recvData->str[inv[j]] = 1;
	}
	
/* 	if (srceArray->decomp->rank < destArray->decomp->rank) { */
/* 	  tmp1 = (int*)calloc(srceArray->decomp->rank, sizeof(int)); */
/* 	  tmp2 = (int*)calloc(srceArray->decomp->rank, sizeof(int)); */
/* 	  for (j = 0; j < srceArray->decomp->rank; ++j) { */
/* 	    tmp1[j] = recvData->str[map[j]]; */
/* 	    tmp2[j] = recvData->numelem[map[j]]; */
/* 	  } */
/* 	  for (j = 0; j < srceArray->decomp->rank; ++j) { */
/* 	    recvData->str[j] = tmp1[j]; */
/* 	    recvData->numelem[j] = tmp2[j]; */
/* 	  } */
/* 	  free(tmp1); */
/* 	  free(tmp2); */
/* 	  for (j = srceArray->decomp->rank; j < rank; ++j) { */
/* 	    recvData->str[j] = recvData->numelem[j] = 1; */
/* 	  } */
/* 	} */

	recvData->startPosn = compute_offset(destArray, destSubregionLocal);
	parti_free_region(destSubregionLocal);
	
	sched->rData[blocks[i]->task] = recvData;
	sched->rMsgSz[blocks[i]->task] = totalSize;
#if PARTI_DEBUG > 1
	print_sched_data("RECV SCHEDULE", recvData);
	printf("totalSize=%d\n", totalSize);
#endif
      }
      parti_free_region(srceSubregionLocal);
    }
    free(blocks);
  }
  
  parti_free_region(srceRegionLocal);
  parti_free_region(destRegionLocal);
  free(inv);
  free(map);
  parti_free_region(srceRegionCopy);
  parti_free_region(destRegionCopy);

  /* save the generated schedule in the hash table */
#if PARTI_DEBUG > 2
  printf("Node %d: saving subarray schedule\n", PARTI_myproc());
#endif
  insert_sub_table(sched, srceArray, destArray, min(srceArray->decomp->rank, destArray->decomp->rank), 
		   dimensionMap, srceRegion->lower, srceRegion->upper, srceRegion->stride, 
		   dimensionMap, destRegion->lower, destRegion->upper, destRegion->stride);
  return sched;
}
