/* 
 * Auth: Christian Hansen (chansen@cs.umd.edu)
 *       University of Maryland, College Park
 * Desc: A couple tests/examples of base MBP functionality
 */

#include "ctest.h"
#include "utilities.h"
#include "birreg.h"
#include "schedules.h"
#include <multiblockparti.h>
#include <assert.h>
#include <stdlib.h>

/*
 * parti_bisect
 */
void parti_bisect(parti_decomp_t* decomp, int rank, int* size, int dim, int* blk) {
  int cnt = 2;
  int idx[2];
  
  if (dim < rank) {
    idx[0] = size[dim]/2; idx[1] = size[dim];
    parti_partition(decomp, dim, blk, cnt, idx);
    blk[dim] = 0;
    parti_bisect(decomp, rank, size, dim+1, blk);
    blk[dim] = 1;
    parti_bisect(decomp, rank, size, dim+1, blk);
  }
}

/* 
 * parti_bisect_alt
 */
void parti_bisect_alt(parti_decomp_t* decomp) {
  int i, *dim, *cnt, **idx;

  dim = (int*)calloc(decomp->rank, sizeof(int));
  cnt = (int*)calloc(decomp->rank, sizeof(int));
  idx = (int**)calloc(decomp->rank, sizeof(int*));
  for (i = 0; i < decomp->rank; ++i) {
    dim[i] = i;
    cnt[i] = 2;
    idx[i] = (int*)calloc(2, sizeof(int)); 
    idx[i][0] = decomp->size[i]/2; idx[i][1] = decomp->size[i];
  }
  
  parti_section(decomp, dim, cnt, idx);

  for (i = 0; i < decomp->rank; ++i) {
    free(idx[i]);
  }
  free(idx);
  free(cnt);
  free(dim);
}

/* 
 * verify_move
 */
int verify_move(parti_darray_t* darray, int* A, int fill[10]) {
  int i, j, k, task, sts = 1;
  parti_region_t* allGhost, *oneGhost;
  decomp_block_t** blocks = 0;
  int count = 0;
  
  task = PARTI_myproc();

  /* get a local bounding box */
  allGhost = parti_alloc_region(3);
  for (j = 0; j < 3; ++j) {
    if (fill[j] < 0) {
      allGhost->lower[j] = parti_global_lower_bound(darray, j) + fill[j];
      allGhost->upper[j] = parti_global_lower_bound(darray, j) + fill[j];
    } else if (fill[j] > 0) {
      allGhost->lower[j] = parti_global_upper_bound(darray, j) + fill[j];
      allGhost->upper[j] = parti_global_upper_bound(darray, j) + fill[j];
    } else {
      allGhost->lower[j] = parti_global_lower_bound(darray, j);
      allGhost->upper[j] = parti_global_upper_bound(darray, j);
    }
    if (allGhost->lower[j] < 0 || allGhost->upper[j] >= 10)
      return 1;
    allGhost->lower[j] = parti_global_to_local(darray, allGhost->lower[j], j, 1);
    allGhost->upper[j] = parti_global_to_local(darray, allGhost->upper[j], j, 1);
  }
  parti_local_size(darray, allGhost->stride);

  /* look at each element and insure that it has changed to the value at the source */
  oneGhost = parti_alloc_region(3);
  for (i = allGhost->lower[0]; i <= allGhost->upper[0]; ++i) {
    oneGhost->lower[0] = oneGhost->upper[0] = parti_local_to_global(darray, i, 0, 1);
    for (j = allGhost->lower[1]; j <= allGhost->upper[1]; ++j) {
      oneGhost->lower[1] = oneGhost->upper[1] = parti_local_to_global(darray, j, 1, 1);
      for (k = allGhost->lower[2]; k <= allGhost->upper[2]; ++k) {
	oneGhost->lower[2] = oneGhost->upper[2] = parti_local_to_global(darray, k, 2, 1);
#ifdef PARTI_DEBUG
 	print_region(oneGhost); 
#endif
	get_region_blocks(oneGhost, &(darray->decomp->root), &blocks, &count);
#ifdef PARTI_DEBUG
	printf("block count:%d\n", count);
#endif
	assert(count == 1); /* one ghost cell should be owned by one and only one task */
#ifdef PARTI_DEBUG
	printf("actual:%d expected:%d\n", 
	       *(A+i*allGhost->stride[1]*allGhost->stride[2]+j*allGhost->stride[2]+k),
	       (*blocks)->task);
#endif
	if (*(A+i*allGhost->stride[1]*allGhost->stride[2]+j*allGhost->stride[2]+k) != (*blocks)->task) {
	  free(blocks);
	  blocks = 0;
	  sts = 0;
	  break;
	} else {
	  free(blocks);
	  blocks = 0;
	  count = 0;
	}
      }
    }
  }
  parti_free_region(oneGhost);
  parti_free_region(allGhost);

  return sts;
}

/* 
 * inter_test
 */
int inter_test(FILE* fp, int n, int p) {
  parti_decomp_t* decomp;
  parti_darray_t* darray;
  int sts;
  
  int rank = 3;
  int size[3], igc[3];
  
  int blk[3], cnt;

  int* A; /* local portion of the array */
  
  parti_sched_t* sched;
  int fill[3];

  size[0] = size[1] = size[2] = 10;
  decomp = parti_create_idecomp(rank, size);
/*   parti_bisect(decomp, rank, size, 0, blk); */
  parti_bisect_alt(decomp);

  igc[0] = igc[1] = igc[2] = 1;
  darray = parti_ialign(decomp, igc, igc);
/*   print_darray(darray); */
  
  cnt = parti_local_size(darray, size);
  A = (int*)malloc(cnt*sizeof(int));
  sts = PARTI_myproc();
  fprintf(fp, "PROC: %d\n", sts);
  elemset('i', A, sts, 3, size, 0, 0);
/*   fprint_array(fp, 'i', A, 3, size, 0, 0); */
  
/*   for (dim = 0; dim < 3; ++dim) { */
/*     printf("LS[%d]:(%d)\n", dim, size[dim]); */
/*     printf("GR[%d]:(%d->%d)\n", dim,  */
/* 	   parti_global_lower_bound(darray, dim),  */
/* 	   parti_global_upper_bound(darray, dim));     */
/*     printf("LR[%d]:(%d->%d)\n", dim,  */
/* 	   parti_local_loop_bound(darray, dim, 0, 1), */
/* 	   parti_local_loop_bound(darray, dim, 9, 1)); */
/*   } */
  
  fill[0] = 0;
  fill[1] = 1;
  fill[2] = 0;
  sched = parti_ghost_sched_vector(darray, fill);
  parti_data_move_int(A, sched, A);
  fprint_array(fp, 'i', A, 3, size, 0, 0);  
  if (!verify_move(darray, A, fill)) 
    printf("verification failed\n");
  else
    printf("verification succeeded\n");
  
  elemset('i', A, sts, 3, size, 0, 0); 
  fill[0] = 1;
  fill[1] = 1;
  fill[2] = 1;
  sched = parti_ghost_sched_span(darray, fill);
  parti_data_move_int(A, sched, A);
  fprint_array(fp, 'i', A, 3, size, 0, 0);
  
  elemset('i', A, sts, 3, size, 0, 0); 
  sched = parti_ghost_sched_all(darray);
  parti_data_move_int(A, sched, A);
  fprint_array(fp, 'i', A, 3, size, 0, 0);
  
  elemset('i', A, sts, 3, size, 0, 0); 
  sched = parti_ghost_sched_abs(darray, 1, 1);
  parti_data_move_int(A, sched, A);
  fprint_array(fp, 'i', A, 3, size, 0, 0);

  parti_free_darray(darray);
  parti_free_decomp(decomp);
  
  return 0;
}

/* 
 * intra_test
 */
int intra_test(FILE* fp, int n, int p) {
  int i;
  
  parti_decomp_t* decomp1, *decomp2;
  parti_darray_t* darray1, *darray2;
  int sts;
  
  int rank = 3;
  int size[3], igc[3];
  
  int blk[3], cnt;

  int* A, *B; /* local portion of the arrays */
  
  int map[3];
  parti_region_t* region1, *region2;
  parti_sched_t* sched;

  size[0] = size[1] = size[2] = 10;
  decomp1 = parti_create_idecomp(rank, size);
/*   parti_bisect(decomp1, rank, size, 0, blk); */
  parti_bisect_alt(decomp1);
  decomp2 = parti_create_idecomp(rank, size);
/*   parti_bisect(decomp2, rank, size, 0, blk); */
  parti_bisect_alt(decomp2);

  igc[0] = igc[1] = igc[2] = 1;
  darray1 = parti_ialign(decomp1, igc, igc);
  darray2 = parti_ialign(decomp2, igc, igc);
  
  sts = PARTI_myproc();
  fprintf(fp, "PROC: %d\n", sts);

  cnt = parti_local_size(darray1, size);
  A = (int*)malloc(cnt*sizeof(int));
  elemset('i', A, sts, 3, size, 0, 0);
  cnt = parti_local_size(darray2, size);
  B = (int*)malloc(cnt*sizeof(int));
  elemset('i', B, sts+10, 3, size, 0, 0);

  map[0] = 0; map[1] = 1; map[2] = 2;
  region1 = parti_alloc_region(rank);
  region2 = parti_alloc_region(rank);

  for (i = 0; i < rank; ++i) {
    region1->lower[i] = 0;
    region1->upper[i] = 2;
    region1->stride[i] = 1;
    region2->lower[i] = 7;
    region2->upper[i] = 9;
    region2->stride[i] = 1;
  }
  sched = parti_region_sched(darray1, region1, darray2, region2, map);
  parti_data_move_int(A, sched, B);
  fprint_array(fp, 'i', A, 3, size, 0, 0);
  fprint_array(fp, 'i', B, 3, size, 0, 0);

  for (i = 0; i < rank; ++i) {
    region1->lower[i] = 4;
    region1->upper[i] = 6;
    region1->stride[i] = 1;
    region2->lower[i] = 0;
    region2->upper[i] = 2;
    region2->stride[i] = 1;
  }
  sched = parti_region_sched(darray2, region2, darray1, region1, map);
  parti_data_move_int(B, sched, A);
  fprint_array(fp, 'i', A, 3, size, 0, 0);
  fprint_array(fp, 'i', B, 3, size, 0, 0);

  parti_free_region(region1);
  parti_free_region(region2);

  parti_free_darray(darray1);
  parti_free_darray(darray2);
  parti_free_decomp(decomp1);
  parti_free_decomp(decomp2);
  
  return 0;
}

void parti_test(FILE* ifp, FILE* ofp, int p, int r) {
  if (p < 8) {
    fprintf(ofp, "need 8 tasks for these tests\n");
  } else { 
    inter_test(ofp, p, r);
    intra_test(ofp, p, r);
  }
}
